# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np


def check_state_back_to_orig():
    np.testing.assert_equal(len(ift.random._rng),1)
    np.testing.assert_equal(len(ift.random._sseq),1)


def test_rand1():
    with ift.random.Context(31):
        a = ift.random.current_rng().integers(0,1000000000)
    with ift.random.Context(31):
        b = ift.random.current_rng().integers(0,1000000000)
    check_state_back_to_orig()
    np.testing.assert_equal(a,b)


def test_rand2():
    sseq = ift.random.spawn_sseq(10)
    with ift.random.Context(sseq[2]):
        a = ift.random.current_rng().integers(0,1000000000)
    with ift.random.Context(sseq[2]):
        b = ift.random.current_rng().integers(0,1000000000)
    check_state_back_to_orig()
    np.testing.assert_equal(a,b)


def test_rand3():
    with ift.random.Context(31):
        sseq = ift.random.spawn_sseq(10)
        with ift.random.Context(sseq[2]):
            a = ift.random.current_rng().integers(0,1000000000)
    with ift.random.Context(31):
        sseq = ift.random.spawn_sseq(1)
        sseq = ift.random.spawn_sseq(1)
        sseq = ift.random.spawn_sseq(1)
        with ift.random.Context(sseq[0]):
            b = ift.random.current_rng().integers(0,1000000000)
    check_state_back_to_orig()
    np.testing.assert_equal(a,b)


def test_rand4():
    with ift.random.Context(31):
        a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(31):
            b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)


def test_rand5():
    ift.random.push_sseq_from_seed(31)
    a = ift.random.current_rng().integers(0,1000000000)
    ift.random.push_sseq_from_seed(31)
    b = ift.random.current_rng().integers(0,1000000000)
    c = ift.random.current_rng().integers(0,1000000000)
    ift.random.pop_sseq()
    d = ift.random.current_rng().integers(0,1000000000)
    ift.random.pop_sseq()
    check_state_back_to_orig()
    np.testing.assert_equal(a,b)
    np.testing.assert_equal(c,d)


def test_rand5b():
    with ift.random.Context(31):
        a = ift.random.current_rng().integers(0,1000000000)
        with ift.random.Context(31):
            b = ift.random.current_rng().integers(0,1000000000)
            c = ift.random.current_rng().integers(0,1000000000)
        d = ift.random.current_rng().integers(0,1000000000)
    check_state_back_to_orig()
    np.testing.assert_equal(a,b)
    np.testing.assert_equal(c,d)


def test_rand6():
    ift.random.push_sseq_from_seed(31)
    state = ift.random.getState()
    a = ift.random.current_rng().integers(0,1000000000)
    ift.random.setState(state)
    b = ift.random.current_rng().integers(0,1000000000)
    np.testing.assert_equal(a,b)
    ift.random.pop_sseq()
    check_state_back_to_orig()


def test_rand6b():
    with ift.random.Context(31):
        state = ift.random.getState()
        a = ift.random.current_rng().integers(0,1000000000)
        ift.random.setState(state)
        b = ift.random.current_rng().integers(0,1000000000)
        np.testing.assert_equal(a,b)
    check_state_back_to_orig()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020-2021 Max-Planck-Society
# Author: Martin Reinecke
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np


def _sqrt_helper(v):
    tmp = np.sqrt(v)
    return (tmp, 0.5/tmp)


def _sinc_helper(v):
    fv = np.sinc(v)
    df = np.empty(v.shape, dtype=v.dtype)
    sel = v != 0.
    v = v[sel]
    df[sel] = (np.cos(np.pi*v)-fv[sel])/v
    df[~sel] = 0
    return (fv, df)


def _expm1_helper(v):
    tmp = np.expm1(v)
    return (tmp, tmp+1.)


def _tanh_helper(v):
    tmp = np.tanh(v)
    return (tmp, 1.-tmp**2)


def _sigmoid_helper(v):
    tmp = np.tanh(v)
    tmp2 = 0.5+(0.5*tmp)
    return (tmp2, 0.5-0.5*tmp**2)


def _reciprocal_helper(v):
    tmp = 1./v
    return (tmp, -tmp**2)


def _abs_helper(v):
    if np.issubdtype(v.dtype, np.complexfloating):
        raise TypeError("Argument must not be complex because abs(z) is not holomorphic")
    return (np.abs(v), np.where(v == 0, np.nan, np.sign(v)))


def _sign_helper(v):
    if np.issubdtype(v.dtype, np.complexfloating):
        raise TypeError("Argument must not be complex")
    return (np.sign(v), np.where(v == 0, np.nan, 0))


def _power_helper(v, expo):
    return (np.power(v, expo), expo*np.power(v, expo-1))


def _clip_helper(v, a_min, a_max):
    if np.issubdtype(v.dtype, np.complexfloating):
        raise TypeError("Argument must not be complex")
    tmp = np.clip(v, a_min, a_max)
    tmp2 = np.ones(v.shape)
    if a_min is not None:
        tmp2 = np.where(tmp == a_min, 0., tmp2)
    if a_max is not None:
        tmp2 = np.where(tmp == a_max, 0., tmp2)
    return (tmp, tmp2)

def _step_helper(v, grad):
    if np.issubdtype(v.dtype, np.complexfloating):
        raise TypeError("Argument must not be complex")
    r = np.zeros(v.shape)
    r[v>=0.] = 1.
    if grad:
        return (r, np.zeros(v.shape))
    return r

def softplus(v):
    fv = np.empty(v.shape, dtype=np.float64 if np.isrealobj(v) else np.complex128)
    selp = v > 33
    selm = v < -33
    sel0 = ~np.logical_or(selp, selm)
    fv[selp] = v[selp]
    fv[sel0] = np.log(1+np.exp(v[sel0]))
    fv[selm] = 0
    return fv


def _softplus_helper(v):
    dtype = np.float64 if np.isrealobj(v) else np.complex128
    fv = np.empty(v.shape, dtype=dtype)
    dfv = np.empty(v.shape, dtype=dtype)
    selp = 33 < v
    selm = v < -33
    sel0 = ~np.logical_or(selp, selm)
    fv[selp] = v[selp]
    fv[sel0] = np.log(1+np.exp(v[sel0]))
    fv[selm] = 0
    dfv[selp] = 1
    dfv[sel0] = 1/(1+np.exp(-v[sel0]))
    dfv[selm] = 0
    return fv, dfv


def exponentiate(v, base):
    return np.power(base, v)


def _exponentiate_helper(v, base):
    tmp = np.power(base, v)
    return (tmp, np.log(base) * tmp)


ptw_dict = {
    "sqrt": (np.sqrt, _sqrt_helper),
    "sin": (np.sin, lambda v: (np.sin(v), np.cos(v))),
    "cos": (np.cos, lambda v: (np.cos(v), -np.sin(v))),
    "tan": (np.tan, lambda v: (np.tan(v), 1./np.cos(v)**2)),
    "sinc": (np.sinc, _sinc_helper),
    "exp": (np.exp, lambda v: (2*(np.exp(v),))),
    "expm1": (np.expm1, _expm1_helper),
    "log": (np.log, lambda v: (np.log(v), 1./v)),
    "log10": (np.log10, lambda v: (np.log10(v), (1./np.log(10.))/v)),
    "log1p": (np.log1p, lambda v: (np.log1p(v), 1./(1.+v))),
    "sinh": (np.sinh, lambda v: (np.sinh(v), np.cosh(v))),
    "cosh": (np.cosh, lambda v: (np.cosh(v), np.sinh(v))),
    "tanh": (np.tanh, _tanh_helper),
    "sigmoid": (lambda v: 0.5+(0.5*np.tanh(v)), _sigmoid_helper),
    "reciprocal": (lambda v: 1./v, _reciprocal_helper),
    "abs": (np.abs, _abs_helper),
    "absolute": (np.abs, _abs_helper),
    "sign": (np.sign, _sign_helper),
    "power": (np.power, _power_helper),
    "clip": (np.clip, _clip_helper),
    "softplus": (softplus, _softplus_helper),
    "exponentiate": (exponentiate, _exponentiate_helper),
    "arctan": (np.arctan, lambda v: (np.arctan(v), 1./(1.+v**2))),
    "unitstep": (lambda v: _step_helper(v, False), lambda v: _step_helper(v, True))
    }


def sigmoid_j(v):
    from jax import numpy as jnp

    # NOTE, the sigmoid used in NIFTy is different to the one commonly referred
    # to as sigmoid in most of the literature.
    return 0.5 + (0.5 * jnp.tanh(v))


def exponentiate_j(v, base):
    from jax import numpy as jnp

    return jnp.power(base, v)


ptw_nifty2jax_dict = {
    "sigmoid": sigmoid_j,
    "exponentiate": exponentiate_j,
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal, assert_raises

from .common import setup_function, teardown_function

pmp = pytest.mark.parametrize
SPACES = [ift.RGSpace((4,)), ift.RGSpace((5))]
SPACE_COMBINATIONS = [(), SPACES[0], SPACES[1], SPACES]


@pmp('domain', SPACE_COMBINATIONS)
@pmp('attribute_desired_type',
     [['domain', ift.DomainTuple], ['val', np.ndarray],
      ['shape', tuple], ['size', (int, np.int64)]])
def test_return_types(domain, attribute_desired_type):
    attribute = attribute_desired_type[0]
    desired_type = attribute_desired_type[1]
    f = ift.Field.full(domain, 1.)
    assert_equal(isinstance(getattr(f, attribute), desired_type), True)


def _spec1(k):
    return 42/(1. + k)**2


def _spec2(k):
    return 42/(1. + k)**3


@pmp('space1', [ift.RGSpace((8,), harmonic=True),
                ift.RGSpace((8, 8), harmonic=True, distances=0.123)])
@pmp('space2', [ift.RGSpace((8,), harmonic=True), ift.LMSpace(12)])
def test_power_synthesize_analyze(space1, space2):
    p1 = ift.PowerSpace(space1)
    fp1 = ift.PS_field(p1, _spec1)
    p2 = ift.PowerSpace(space2)
    fp2 = ift.PS_field(p2, _spec2)
    op1 = ift.create_power_operator((space1, space2), _spec1, 0, float)
    op2 = ift.create_power_operator((space1, space2), _spec2, 1, float)
    opfull = op2 @ op1

    samples = 120
    sc1 = ift.StatCalculator()
    sc2 = ift.StatCalculator()
    for ii in range(samples):
        sk = opfull.draw_sample()
        sp = ift.power_analyze(sk, spaces=(0, 1), keep_phase_information=False)
        sc1.add(sp.sum(spaces=1)/fp2.s_sum())
        sc2.add(sp.sum(spaces=0)/fp1.s_sum())
    assert_allclose(sc1.mean.val, fp1.val, rtol=0.2)
    assert_allclose(sc2.mean.val, fp2.val, rtol=0.2)


@pmp('space1', [
    ift.RGSpace((8,), harmonic=True),
    ift.RGSpace((8, 8), harmonic=True, distances=0.123)
])
@pmp('space2', [ift.RGSpace((8,), harmonic=True), ift.LMSpace(12)])
def test_DiagonalOperator_power_analyze2(space1, space2):
    fp1 = ift.PS_field(ift.PowerSpace(space1), _spec1)
    fp2 = ift.PS_field(ift.PowerSpace(space2), _spec2)

    S_1 = ift.create_power_operator((space1, space2), _spec1, 0, float)
    S_2 = ift.create_power_operator((space1, space2), _spec2, 1, float)
    S_full = S_2 @ S_1

    samples = 500
    sc1 = ift.StatCalculator()
    sc2 = ift.StatCalculator()

    for ii in range(samples):
        sk = S_full.draw_sample()
        sp = ift.power_analyze(sk, spaces=(0, 1), keep_phase_information=False)
        sc1.add(sp.sum(spaces=1)/fp2.s_sum())
        sc2.add(sp.sum(spaces=0)/fp1.s_sum())

    assert_allclose(sc1.mean.val, fp1.val, rtol=0.2)
    assert_allclose(sc2.mean.val, fp2.val, rtol=0.2)


@pmp('space', [
    ift.RGSpace((8,), harmonic=True), (),
    ift.RGSpace((8, 8), harmonic=True, distances=0.123),
    ift.RGSpace((2, 3, 7))
])
def test_norm(space):
    f = ift.Field.from_random(domain=space, random_type="normal", dtype=np.complex128)
    gd = f.val.reshape(-1)
    assert_allclose(f.norm(), np.linalg.norm(gd))
    assert_allclose(f.norm(1), np.linalg.norm(gd, ord=1))
    assert_allclose(f.norm(2), np.linalg.norm(gd, ord=2))
    assert_allclose(f.norm(3), np.linalg.norm(gd, ord=3))
    assert_allclose(f.norm(np.inf), np.linalg.norm(gd, ord=np.inf))


def test_vdot():
    s = ift.RGSpace((10,))
    f1 = ift.Field.from_random(domain=s, random_type="normal", dtype=np.complex128)
    f2 = ift.Field.from_random(domain=s, random_type="normal", dtype=np.complex128)
    assert_allclose(f1.s_vdot(f2), f1.vdot(f2, spaces=0).val)
    assert_allclose(f1.s_vdot(f2), np.conj(f2.s_vdot(f1)))


def test_vdot2():
    x1 = ift.RGSpace((200,))
    x2 = ift.RGSpace((150,))
    m = ift.Field.full((x1, x2), .5)
    res = m.vdot(m, spaces=1)
    assert_allclose(res.val, 37.5)


def test_outer():
    x1 = ift.RGSpace((9,))
    x2 = ift.RGSpace((3,))
    m1 = ift.Field.full(x1, .5)
    m2 = ift.Field.full(x2, 3.)
    res = m1.outer(m2)
    assert_allclose(res.val, np.full((9, 3), 1.5))


def test_sum():
    x1 = ift.RGSpace((9,), distances=2.)
    x2 = ift.RGSpace((2, 12), distances=(0.3,))
    m1 = ift.Field(ift.makeDomain(x1), np.arange(9))
    m2 = ift.Field.full(ift.makeDomain((x1, x2)), 0.45)
    res1 = m1.s_sum()
    res2 = m2.sum(spaces=1)
    assert_allclose(res1, 36)
    assert_allclose(res2.val, np.full(9, 2*12*0.45))


def test_integrate():
    x1 = ift.RGSpace((9,), distances=2.)
    x2 = ift.RGSpace((2, 12), distances=(0.3,))
    m1 = ift.Field(ift.makeDomain(x1), np.arange(9))
    m2 = ift.Field.full(ift.makeDomain((x1, x2)), 0.45)
    res1 = m1.s_integrate()
    res2 = m2.integrate(spaces=1)
    assert_allclose(res1, 36*2)
    assert_allclose(res2.val, np.full(9, 2*12*0.45*0.3**2))
    for m in [m1, m2]:
        res3 = m.integrate()
        res4 = m.s_integrate()
        assert_allclose(res3.val, res4)
    dom = ift.HPSpace(3)
    assert_allclose(ift.full(dom, 1).s_integrate(), 4*np.pi)


def test_dataconv():
    s1 = ift.RGSpace((10,))
    gd = np.arange(s1.shape[0])
    assert_equal(gd, ift.makeField(s1, gd).val)


def test_cast_domain():
    s1 = ift.RGSpace((10,))
    s2 = ift.RGSpace((10,), distances=20.)
    d = np.arange(s1.shape[0])
    d2 = ift.makeField(s1, d).cast_domain(s2).val
    assert_equal(d, d2)


def test_empty_domain():
    f = ift.Field.full((), 5)
    assert_equal(f.val, 5)
    f = ift.Field.full(None, 5)
    assert_equal(f.val, 5)


def test_trivialities():
    s1 = ift.RGSpace((10,))
    f1 = ift.Field.full(s1, 27)
    assert_equal(f1.clip(a_min=29, a_max=50).val, 29.)
    assert_equal(f1.clip(a_min=0, a_max=25).val, 25.)
    assert_equal(f1.val, f1.real.val)
    assert_equal(f1.val, (+f1).val)
    f1 = ift.Field.full(s1, 27. + 3j)
    assert_equal(f1.ptw("reciprocal").val, (1./f1).val)
    assert_equal(f1.real.val, 27.)
    assert_equal(f1.imag.val, 3.)
    assert_equal(f1.s_sum(), f1.sum(0).val)
    assert_equal(f1.conjugate().val,
                 ift.Field.full(s1, 27. - 3j).val)
    f1 = ift.makeField(s1, np.arange(10))
    # assert_equal(f1.min(), 0)
    # assert_equal(f1.max(), 9)
    assert_equal(f1.s_prod(), 0)


def test_weight():
    s1 = ift.RGSpace((10,))
    f = ift.Field.full(s1, 10.)
    f2 = f.weight(1)
    assert_equal(f.weight(1).val, f2.val)
    assert_equal(f.domain.total_volume(), 1)
    assert_equal(f.domain.total_volume(0), 1)
    assert_equal(f.domain.total_volume((0,)), 1)
    assert_equal(f.total_volume(), 1)
    assert_equal(f.total_volume(0), 1)
    assert_equal(f.total_volume((0,)), 1)
    assert_equal(f.domain.scalar_weight(), 0.1)
    assert_equal(f.domain.scalar_weight(0), 0.1)
    assert_equal(f.domain.scalar_weight((0,)), 0.1)
    assert_equal(f.scalar_weight(), 0.1)
    assert_equal(f.scalar_weight(0), 0.1)
    assert_equal(f.scalar_weight((0,)), 0.1)
    s1 = ift.GLSpace(10)
    f = ift.Field.full(s1, 10.)
    assert_equal(f.domain.scalar_weight(), None)
    assert_equal(f.domain.scalar_weight(0), None)
    assert_equal(f.domain.scalar_weight((0,)), None)


@pmp('dom', [ift.RGSpace(10), ift.GLSpace(10)])
@pmp('dt', [np.float64, np.complex128])
def test_reduction(dom, dt):
    s1 = ift.Field.full(dom, dt(1.))
    assert_allclose(s1.s_mean(), 1.)
    assert_allclose(s1.mean(0).val, 1.)
    assert_allclose(s1.s_var(), 0., atol=1e-14)
    assert_allclose(s1.var(0).val, 0., atol=1e-14)
    assert_allclose(s1.s_std(), 0., atol=1e-14)
    assert_allclose(s1.std(0).val, 0., atol=1e-14)


def test_err():
    s1 = ift.RGSpace((10,))
    s2 = ift.RGSpace((11,))
    f1 = ift.Field.full(s1, 27)
    with assert_raises(ValueError):
        f2 = ift.Field(ift.DomainTuple.make(s2), f1.val)
    with assert_raises(TypeError):
        f2 = ift.Field.full(s2, "xyz")
    with assert_raises(TypeError):
        if f1:
            pass
    with assert_raises(TypeError):
        f1.full((2, 4, 6))
    with assert_raises(TypeError):
        f2 = ift.Field(None, None)
    with assert_raises(TypeError):
        f2 = ift.Field(s1, None)
    with assert_raises(ValueError):
        f1.imag
    with assert_raises(TypeError):
        f1.vdot(42)
    with assert_raises(ValueError):
        f1.vdot(ift.Field.full(s2, 1.))
    with assert_raises(TypeError):
        ift.full(s1, [2, 3])
    with assert_raises(TypeError):
        ift.Field(s2, [0, 1])
    with assert_raises(TypeError):
        f1.outer([0, 1])
    with assert_raises(ValueError):
        f1.extract(s2)
    with assert_raises(TypeError):
        f1 += f1
    f2 = ift.Field.full(s2, 27)
    with assert_raises(ValueError):
        f1 + f2


def test_stdfunc():
    s = ift.RGSpace((200,))
    f = ift.Field.full(s, 27)
    assert_equal(f.val, 27)
    assert_equal(f.shape, (200,))
    assert_equal(f.dtype, np.int64)
    fx = ift.full(f.domain, 0)
    assert_equal(f.dtype, fx.dtype)
    assert_equal(f.shape, fx.shape)
    assert_equal(fx.val, 0)
    fx = ift.full(f.domain, 1)
    assert_equal(f.dtype, fx.dtype)
    assert_equal(f.shape, fx.shape)
    assert_equal(fx.val, 1)
    fx = ift.full(f.domain, 67.)
    assert_equal(f.shape, fx.shape)
    assert_equal(fx.val, 67.)
    f = ift.Field.from_random(s, "normal")
    f2 = ift.Field.from_random(s, "normal")
    assert_equal((f > f2).val, f.val > f2.val)
    assert_equal((f >= f2).val, f.val >= f2.val)
    assert_equal((f < f2).val, f.val < f2.val)
    assert_equal((f <= f2).val, f.val <= f2.val)
    assert_equal((f != f2).val, f.val != f2.val)
    assert_equal((f == f2).val, f.val == f2.val)
    assert_equal((f + f2).val, f.val + f2.val)
    assert_equal((f - f2).val, f.val - f2.val)
    assert_equal((f*f2).val, f.val*f2.val)
    assert_equal((f/f2).val, f.val/f2.val)
    assert_equal((-f).val, -(f.val))
    assert_equal(abs(f).val, abs(f.val))


def test_emptydomain():
    f = ift.Field.full((), 3.)
    assert_equal(f.s_sum(), 3.)
    assert_equal(f.s_prod(), 3.)
    assert_equal(f.val, 3.)
    assert_equal(f.val.shape, ())
    assert_equal(f.val.size, 1)
    assert_equal(f.s_vdot(f), 9.)


@pmp('num', [float(5), 5.])
@pmp('dom', [ift.RGSpace((8,), harmonic=True), ()])
@pmp('func', [
    "exp", "log", "sin", "cos", "tan", "sinh", "cosh", "sinc", "absolute",
    "sign", "log10", "log1p", "expm1"
])
def test_funcs(num, dom, func):
    num = 5
    f = ift.Field.full(dom, num)
    res = f.ptw(func)
    res2 = getattr(np, func)(num)
    assert_allclose(res.val, res2)


@pmp('rtype', ['normal', 'pm1', 'uniform'])
@pmp('dtype', [np.float64, np.complex128])
def test_from_random(rtype, dtype):
    sp = ift.RGSpace(3)
    ift.Field.from_random(sp, rtype, dtype=dtype)


def test_field_of_objects():
    arr = np.array(['x', 'y', 'z'])
    sp = ift.RGSpace(3)
    with assert_raises(TypeError):
        ift.Field(sp, arr)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce

import numpy as np

from ..field import Field
from .structured_domain import StructuredDomain


class RGSpace(StructuredDomain):
    """Represents a regular Cartesian grid.

    Parameters
    ----------
    shape : int or tuple of int
        Number of grid points or numbers of gridpoints along each axis.
    distances : None or float or tuple of float, optional
        Distance between two grid points along each axis.

        By default (distances=None):
          - If harmonic==True, all distances will be set to 1
          - If harmonic==False, the distance along each axis will be
            set to the inverse of the number of points along that axis.

    harmonic : bool, optional
        Whether the space represents a grid in position or harmonic space.
        Default: False.

    Notes
    -----
    Topologically, a n-dimensional RGSpace is a n-Torus, i.e. it has periodic
    boundary conditions.
    """
    _needed_for_hash = ["_rdistances", "_shape", "_harmonic"]

    def __init__(self, shape, distances=None, harmonic=False, _realdistances=None):
        self._harmonic = bool(harmonic)
        if np.isscalar(shape):
            shape = (shape,)
        self._shape = tuple(int(i) for i in shape)
        if min(self._shape) < 0:
            raise ValueError('Negative number of pixels encountered')

        if _realdistances is not None:
            self._rdistances = _realdistances
        else:
            if distances is None:
                self._rdistances = tuple(1. / (np.array(self._shape)))
            elif np.isscalar(distances):
                if self.harmonic:
                    self._rdistances = tuple(
                        1. / (np.array(self._shape) * float(distances)))
                else:
                    self._rdistances = (float(distances),) * len(self._shape)
            else:
                temp = np.empty(len(self.shape), dtype=np.float64)
                temp[:] = distances
                if self._harmonic:
                    temp = 1. / (np.array(self._shape) * temp)
                self._rdistances = tuple(temp)
        self._hdistances = tuple(
            1. / (np.array(self.shape)*np.array(self._rdistances)))
        if min(self._rdistances) <= 0:
            raise ValueError('Non-positive distances encountered')

        self._dvol = float(reduce(lambda x, y: x*y, self.distances))
        self._size = int(reduce(lambda x, y: x*y, self._shape))

    def __repr__(self):
        return ("RGSpace(shape={}, distances={}, harmonic={})"
                .format(self.shape, self.distances, self.harmonic))

    @property
    def harmonic(self):
        return self._harmonic

    @property
    def shape(self):
        return self._shape

    @property
    def size(self):
        return self._size

    @property
    def scalar_dvol(self):
        return self._dvol

    def _get_dist_array(self):
        res = np.arange(self.shape[0], dtype=np.float64)
        res = np.minimum(res, self.shape[0]-res)*self.distances[0]
        if len(self.shape) == 1:
            return Field.from_raw(self, res)
        res *= res
        for i in range(1, len(self.shape)):
            tmp = np.arange(self.shape[i], dtype=np.float64)
            tmp = np.minimum(tmp, self.shape[i]-tmp)*self.distances[i]
            tmp *= tmp
            res = np.add.outer(res, tmp)
        return Field.from_raw(self, np.sqrt(res))

    def get_k_length_array(self):
        if (not self.harmonic):
            raise NotImplementedError
        return self._get_dist_array()

    def get_unique_k_lengths(self):
        if (not self.harmonic):
            raise NotImplementedError
        dimensions = len(self.shape)
        if dimensions == 1:  # extra easy
            maxdist = self.shape[0]//2
            return np.arange(maxdist+1, dtype=np.float64) * self.distances[0]
        if np.all(self.distances == self.distances[0]):  # shortcut
            maxdist = np.asarray(self.shape)//2
            tmp = np.sum(maxdist*maxdist)
            tmp = np.zeros(tmp+1, dtype=bool)
            t2 = np.arange(maxdist[0]+1, dtype=np.int64)
            t2 *= t2
            for i in range(1, dimensions):
                t3 = np.arange(maxdist[i]+1, dtype=np.int64)
                t3 *= t3
                t2 = np.add.outer(t2, t3)
            tmp[t2] = True
            return np.sqrt(np.nonzero(tmp)[0])*self.distances[0]
        else:  # do it the hard way
            tmp = self.get_k_length_array().val
            tmp = np.unique(tmp)
            tol = 1e-12*tmp[-1]
            # remove all points that are closer than tol to their right
            # neighbors.
            # I'm appending the last value*2 to the array to treat the
            # rightmost point correctly.
            return tmp[np.diff(np.r_[tmp, 2*tmp[-1]]) > tol]

    @staticmethod
    def _kernel(x, sigma):
        return (x*x * (-2.*np.pi*np.pi*sigma*sigma)).ptw("exp")

    def get_fft_smoothing_kernel_function(self, sigma):
        if (not self.harmonic):
            raise NotImplementedError
        return lambda x: self._kernel(x, sigma)

    def get_conv_kernel_from_func(self, func):
        """Creates a convolution kernel defined by a function.

        Assumes the function to be radially symmetric, e.g. only dependant on
        distance.

        Parameters
        ----------
        func: function
            This function needs to take exactly one argument, which is
            distance from center (in the same units as the RGSpace distances),
            and return the kernel amplitude at that distance.
        """
        from ..operators.harmonic_operators import HarmonicTransformOperator
        if (not self.harmonic):
            raise NotImplementedError
        op = HarmonicTransformOperator(self, self.get_default_codomain())
        dist = op.target[0]._get_dist_array()
        kernel = Field(op.target, func(dist.val))
        kernel = kernel / kernel.s_integrate()
        return op.adjoint_times(kernel.weight(1))

    def get_default_codomain(self):
        """Returns a :class:`RGSpace` object representing the (position or
        harmonic) partner domain of `self`, depending on `self.harmonic`.

        Returns
        -------
        RGSpace
            The partner domain
        """
        return RGSpace(self.shape, None, not self.harmonic, self._rdistances)

    def check_codomain(self, codomain):
        """Raises `TypeError` if `codomain` is not a matching partner domain
        for `self`.
        """
        if not isinstance(codomain, RGSpace):
            raise TypeError("domain is not a RGSpace")

        if self.shape != codomain.shape:
            raise AttributeError("The shapes of domain and codomain must be "
                                 "identical.")

        if self.harmonic == codomain.harmonic:
            raise AttributeError("domain.harmonic and codomain.harmonic must "
                                 "not be the same.")

        # Check if the distances match, i.e. dist' = 1 / (num * dist)
        if not np.all(abs(np.array(self.shape) *
                          np.array(self.distances) *
                          np.array(codomain.distances)-1) < 1e-7):
            raise AttributeError("The grid-distances of domain and codomain "
                                 "do not match.")

    @property
    def distances(self):
        """tuple of float : Distance between grid points along each axis.
        The n-th entry of the tuple is the distance between neighboring
        grid points along the n-th dimension.
        """
        return self._hdistances if self._harmonic else self._rdistances



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import sys

sys.path.append("demos")

import tempfile

import nifty8 as ift
from getting_started_7_config_file import builder_dct


def test_optimize_kl_cfg_save_and_load():
    cfg = ift.OptimizeKLConfig.from_file("demos/getting_started_7_config_file.cfg", builder_dct)
    with tempfile.NamedTemporaryFile() as f:
        file_name = f.name
        cfg.to_file(file_name)
        cfg0 = ift.OptimizeKLConfig.from_file(file_name, builder_dct)
        assert cfg == cfg0
    # /Move to tests



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

###############################################################################
# Log-normal field reconstruction from Poissonian data with inhomogenous
# exposure (in case for 2D mode)
# 1D (set mode=0), 2D (mode=1), or on the sphere (mode=2)
###############################################################################

import sys

import numpy as np

import nifty8 as ift


def exposure_2d(domain):
    # Structured exposure for 2D mode
    x_shape, y_shape = domain.shape
    exposure = np.ones(domain.shape)
    exposure[x_shape//3:x_shape//2, :] *= 2.
    exposure[x_shape*4//5:x_shape, :] *= .1
    exposure[x_shape//2:x_shape*3//2, :] *= 3.
    exposure[:, x_shape//3:x_shape//2] *= 2.
    exposure[:, x_shape*4//5:x_shape] *= .1
    exposure[:, x_shape//2:x_shape*3//2] *= 3.
    return ift.Field.from_raw(domain, exposure)


def main():
    # Choose space on which the signal field is defined
    if len(sys.argv) == 2:
        mode = int(sys.argv[1])
    else:
        mode = 1

    if mode == 0:
        # One-dimensional regular grid with uniform exposure of 10
        position_space = ift.RGSpace(1024)
        exposure = ift.Field.full(position_space, 10.)
    elif mode == 1:
        # Two-dimensional regular grid with inhomogeneous exposure
        position_space = ift.RGSpace([512, 512])
        exposure = exposure_2d(position_space)
    else:
        # Sphere with uniform exposure of 100
        position_space = ift.HPSpace(128)
        exposure = ift.Field.full(position_space, 100.)

    # Define harmonic space and harmonic transform
    harmonic_space = position_space.get_default_codomain()
    HT = ift.HarmonicTransformOperator(harmonic_space, position_space)

    # Define amplitude (square root of power spectrum)
    def sqrtpspec(k):
        return 1./(20. + k**2)

    p_space = ift.PowerSpace(harmonic_space)
    pd = ift.PowerDistributor(harmonic_space, p_space)
    a = ift.PS_field(p_space, sqrtpspec)
    A = pd(a)

    # Define sky operator
    sky = ift.exp(HT(ift.makeOp(A)).ducktape("domain"))

    M = ift.DiagonalOperator(exposure)
    GR = ift.GeometryRemover(position_space)
    # Define instrumental response
    R = GR(M)

    # Generate mock data and define likelihood energy operator
    d_space = R.target[0]
    lamb = R(sky)
    mock_position = ift.from_random(sky.domain, 'normal')
    data = lamb(mock_position)
    data = ift.random.current_rng().poisson(data.val.astype(np.float64))
    data = ift.Field.from_raw(d_space, data)
    likelihood_energy = ift.PoissonianEnergy(data) @ lamb

    # Settings for minimization
    ic_newton = ift.DeltaEnergyController(
        name='Newton', iteration_limit=100, tol_rel_deltaE=1e-8)
    minimizer = ift.NewtonCG(ic_newton)

    # Compute MAP solution by minimizing the information Hamiltonian
    sl = ift.optimize_kl(likelihood_energy, 1, 0, minimizer, None, None,
                         output_directory="getting_started_2_results",
                         export_operator_outputs={"signal": sky})


if __name__ == '__main__':
    main()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from itertools import combinations

import numpy as np

from .domain_tuple import DomainTuple
from .field import Field
from .linearization import Linearization
from .multi_domain import MultiDomain
from .multi_field import MultiField
from .operators.adder import Adder
from .operators.endomorphic_operator import EndomorphicOperator
from .operators.energy_operators import (EnergyOperator,
                                         LikelihoodEnergyOperator)
from .operators.linear_operator import LinearOperator
from .operators.operator import Operator
from .probing import StatCalculator
from .sugar import from_random, is_fieldlike, is_operator
from .utilities import issingleprec, myassert

__all__ = ["check_linear_operator", "check_operator", "assert_allclose", "minisanity"]


def check_linear_operator(op, domain_dtype=np.float64, target_dtype=np.float64,
                          atol=1e-12, rtol=1e-12, only_r_linear=False):
    """Checks an operator for algebraic consistency of its capabilities.

    Checks whether times(), adjoint_times(), inverse_times() and
    adjoint_inverse_times() (if in capability list) is implemented
    consistently. Additionally, it checks whether the operator is linear.

    Parameters
    ----------
    op : LinearOperator
        Operator which shall be checked.
    domain_dtype : dtype
        The data type of the random vectors in the operator's domain. Default
        is `np.float64`.
    target_dtype : dtype
        The data type of the random vectors in the operator's target. Default
        is `np.float64`.
    atol : float
        Absolute tolerance for the check. If rtol is specified,
        then satisfying any tolerance will let the check pass.
        Default: 0.
    rtol : float
        Relative tolerance for the check. If atol is specified,
        then satisfying any tolerance will let the check pass.
        Default: 0.
    only_r_linear: bool
        set to True if the operator is only R-linear, not C-linear.
        This will relax the adjointness test accordingly.
    """
    if not isinstance(op, LinearOperator):
        raise TypeError('This test tests only linear operators.')
    _domain_check_linear(op, domain_dtype)
    _domain_check_linear(op.adjoint, target_dtype)
    _domain_check_linear(op.inverse, target_dtype)
    _domain_check_linear(op.adjoint.inverse, domain_dtype)
    _purity_check(op, from_random(op.domain, dtype=domain_dtype))
    _purity_check(op.adjoint.inverse, from_random(op.domain, dtype=domain_dtype))
    _purity_check(op.adjoint, from_random(op.target, dtype=target_dtype))
    _purity_check(op.inverse, from_random(op.target, dtype=target_dtype))
    _check_linearity(op, domain_dtype, atol, rtol)
    _check_linearity(op.adjoint, target_dtype, atol, rtol)
    _check_linearity(op.inverse, target_dtype, atol, rtol)
    _check_linearity(op.adjoint.inverse, domain_dtype, atol, rtol)
    _full_implementation(op, domain_dtype, target_dtype, atol, rtol,
                         only_r_linear)
    _full_implementation(op.adjoint, target_dtype, domain_dtype, atol, rtol,
                         only_r_linear)
    _full_implementation(op.inverse, target_dtype, domain_dtype, atol, rtol,
                         only_r_linear)
    _full_implementation(op.adjoint.inverse, domain_dtype, target_dtype, atol,
                         rtol, only_r_linear)
    _check_sqrt(op, domain_dtype)
    _check_sqrt(op.adjoint, target_dtype)
    _check_sqrt(op.inverse, target_dtype)
    _check_sqrt(op.adjoint.inverse, domain_dtype)


def check_operator(op, loc, tol=1e-12, ntries=100, perf_check=True,
                   only_r_differentiable=True, metric_sampling=True):
    """Performs various checks of the implementation of linear and nonlinear
    operators.

    Computes the Jacobian with finite differences and compares it to the
    implemented Jacobian.

    Parameters
    ----------
    op : Operator
        Operator which shall be checked.
    loc : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
        An Field or MultiField instance which has the same domain
        as op. The location at which the gradient is checked
    tol : float
        Tolerance for the check.
    perf_check : Boolean
        Do performance check. May be disabled for very unimportant operators.
    only_r_differentiable : Boolean
        Jacobians of C-differentiable operators need to be C-linear.
        Default: True
    metric_sampling: Boolean
        If op is an EnergyOperator, metric_sampling determines whether the
        test shall try to sample from the metric or not.
    """
    if not isinstance(op, Operator):
        raise TypeError('This test tests only (nonlinear) operators.')
    _domain_check_nonlinear(op, loc)
    _purity_check(op, loc)
    _performance_check(op, loc, bool(perf_check))
    _linearization_value_consistency(op, loc)
    _jac_vs_finite_differences(op, loc, np.sqrt(tol), ntries,
                               only_r_differentiable)
    _check_nontrivial_constant(op, loc, tol, ntries, only_r_differentiable,
                               metric_sampling)
    _check_likelihood_energy(op, loc)


def assert_allclose(f1, f2, atol=0, rtol=1e-7):
    if isinstance(f1, Field):
        return np.testing.assert_allclose(f1.val, f2.val, atol=atol, rtol=rtol)
    if f1.domain is not f2.domain:
        raise AssertionError
    for key, val in f1.items():
        assert_allclose(val, f2[key], atol=atol, rtol=rtol)


def assert_equal(f1, f2):
    if isinstance(f1, Field):
        return np.testing.assert_equal(f1.val, f2.val)
    if f1.domain is not f2.domain:
        raise AssertionError
    for key, val in f1.items():
        assert_equal(val, f2[key])


def _adjoint_implementation(op, domain_dtype, target_dtype, atol, rtol,
                            only_r_linear):
    needed_cap = op.TIMES | op.ADJOINT_TIMES
    if (op.capability & needed_cap) != needed_cap:
        return
    f1 = from_random(op.domain, "normal", dtype=domain_dtype)
    f2 = from_random(op.target, "normal", dtype=target_dtype)
    res1 = f1.s_vdot(op.adjoint_times(f2))
    res2 = op.times(f1).s_vdot(f2)
    if only_r_linear:
        res1, res2 = res1.real, res2.real
    np.testing.assert_allclose(res1, res2, atol=atol, rtol=rtol)


def _inverse_implementation(op, domain_dtype, target_dtype, atol, rtol):
    needed_cap = op.TIMES | op.INVERSE_TIMES
    if (op.capability & needed_cap) != needed_cap:
        return
    foo = from_random(op.target, "normal", dtype=target_dtype)
    res = op(op.inverse_times(foo))
    assert_allclose(res, foo, atol=atol, rtol=rtol)

    foo = from_random(op.domain, "normal", dtype=domain_dtype)
    res = op.inverse_times(op(foo))
    assert_allclose(res, foo, atol=atol, rtol=rtol)


def _full_implementation(op, domain_dtype, target_dtype, atol, rtol,
                         only_r_linear):
    _adjoint_implementation(op, domain_dtype, target_dtype, atol, rtol,
                            only_r_linear)
    _inverse_implementation(op, domain_dtype, target_dtype, atol, rtol)


def _check_linearity(op, domain_dtype, atol, rtol):
    needed_cap = op.TIMES
    if (op.capability & needed_cap) != needed_cap:
        return
    fld1 = from_random(op.domain, "normal", dtype=domain_dtype)
    fld2 = from_random(op.domain, "normal", dtype=domain_dtype)
    alpha = 0.42
    val1 = op(alpha*fld1+fld2)
    val2 = alpha*op(fld1)+op(fld2)
    assert_allclose(val1, val2, atol=atol, rtol=rtol)


def _domain_check_linear(op, domain_dtype=None, inp=None):
    _domain_check(op)
    needed_cap = op.TIMES
    if (op.capability & needed_cap) != needed_cap:
        return
    if domain_dtype is not None:
        inp = from_random(op.domain, "normal", dtype=domain_dtype)
    elif inp is None:
        raise ValueError('Need to specify either dtype or inp')
    myassert(inp.domain is op.domain)
    myassert(op(inp).domain is op.target)


def _check_sqrt(op, domain_dtype):
    if not isinstance(op, EndomorphicOperator):
        try:
            op.get_sqrt()
            raise RuntimeError("Operator implements get_sqrt() although it is not an endomorphic operator.")
        except AttributeError:
            return
    try:
        sqop = op.get_sqrt()
    except (NotImplementedError, ValueError):
        return
    fld = from_random(op.domain, dtype=domain_dtype)
    a = op(fld)
    b = (sqop.adjoint @ sqop)(fld)
    return assert_allclose(a, b, rtol=1e-15)


def _domain_check_nonlinear(op, loc):
    _domain_check(op)
    myassert(isinstance(loc, (Field, MultiField)))
    myassert(loc.domain is op.domain)
    for wm in [False, True]:
        lin = Linearization.make_var(loc, wm)
        reslin = op(lin)
        myassert(lin.domain is op.domain)
        myassert(lin.target is op.domain)
        myassert(lin.val.domain is lin.domain)
        myassert(reslin.domain is op.domain)
        myassert(reslin.target is op.target)
        myassert(reslin.val.domain is reslin.target)
        myassert(reslin.target is op.target)
        myassert(reslin.jac.domain is reslin.domain)
        myassert(reslin.jac.target is reslin.target)
        myassert(lin.want_metric == reslin.want_metric)
        _domain_check_linear(reslin.jac, inp=loc)
        _domain_check_linear(reslin.jac.adjoint, inp=reslin.jac(loc))
        if reslin.metric is not None:
            myassert(reslin.metric.domain is reslin.metric.target)
            myassert(reslin.metric.domain is op.domain)


def _domain_check(op):
    for dd in [op.domain, op.target]:
        if not isinstance(dd, (DomainTuple, MultiDomain)):
            raise TypeError(
                'The domain and the target of an operator need to',
                'be instances of either DomainTuple or MultiDomain.')


def _performance_check(op, pos, raise_on_fail):
    class CountingOp(LinearOperator):
        def __init__(self, domain):
            from .sugar import makeDomain

            self._domain = self._target = makeDomain(domain)
            self._capability = self.TIMES | self.ADJOINT_TIMES
            self._count = 0

        def apply(self, x, mode):
            self._count += 1
            return x

        @property
        def count(self):
            return self._count
    for wm in [False, True]:
        cop = CountingOp(op.domain)
        myop = op @ cop
        myop(pos)
        cond = [cop.count != 1]
        lin = myop(Linearization.make_var(pos, wm))
        cond.append(cop.count != 2)
        lin.jac(pos)
        cond.append(cop.count != 3)
        lin.jac.adjoint(lin.val)
        cond.append(cop.count != 4)
        if lin.metric is not None:
            lin.metric(pos)
            cond.append(cop.count != 6)
        if any(cond):
            s = 'The operator has a performance problem (want_metric={}).'.format(wm)
            from .logger import logger
            logger.error(s)
            logger.info(cond)
            if raise_on_fail:
                raise RuntimeError(s)


def _purity_check(op, pos):
    if isinstance(op, LinearOperator) and (op.capability & op.TIMES) != op.TIMES:
        return
    res0 = op(pos)
    res1 = op(pos)
    assert_equal(res0, res1)


def _get_acceptable_location(op, loc, lin):
    if not np.isfinite(lin.val.s_sum()):
        raise ValueError('Initial value must be finite')
    direction = from_random(loc.domain, dtype=loc.dtype)
    dirder = lin.jac(direction)
    fac = 1e-3 if issingleprec(loc.dtype) else 1e-6
    if dirder.norm() == 0:
        direction = direction * (lin.val.norm() * fac)
    else:
        direction = direction * (lin.val.norm() * fac / dirder.norm())
    # Find a step length that leads to a "reasonable" location
    for i in range(50):
        try:
            loc2 = loc + direction
            lin2 = op(Linearization.make_var(loc2, lin.want_metric))
            if np.isfinite(lin2.val.s_sum()) and abs(lin2.val.s_sum()) < 1e20:
                break
        except FloatingPointError:
            pass
        direction = direction * 0.5
    else:
        raise ValueError("could not find a reasonable initial step")
    return loc2, lin2


def _linearization_value_consistency(op, loc):
    for wm in [False, True]:
        lin = Linearization.make_var(loc, wm)
        fld0 = op(loc)
        fld1 = op(lin).val
        assert_allclose(fld0, fld1, 0, 1e-7)


def _check_nontrivial_constant(op, loc, tol, ntries, only_r_differentiable,
                               metric_sampling):
    if isinstance(op.domain, DomainTuple):
        return
    keys = op.domain.keys()
    combis = []
    if len(keys) > 4:
        from .logger import logger
        logger.warning('Operator domain has more than 4 keys.')
        logger.warning('Check derivatives only with one constant key at a time.')
        combis = [[kk] for kk in keys]
    else:
        for ll in range(1, len(keys)):
            combis.extend(list(combinations(keys, ll)))
    for cstkeys in combis:
        varkeys = set(keys) - set(cstkeys)
        cstloc = loc.extract_by_keys(cstkeys)
        varloc = loc.extract_by_keys(varkeys)

        val0 = op(loc)
        _, op0 = op.simplify_for_constant_input(cstloc)
        myassert(op0.domain is varloc.domain)
        val1 = op0(varloc)
        assert_equal(val0, val1)

        lin = Linearization.make_partial_var(loc, cstkeys, want_metric=True)
        lin0 = Linearization.make_var(varloc, want_metric=True)
        oplin0 = op0(lin0)
        oplin = op(lin)

        myassert(oplin.jac.target is oplin0.jac.target)
        rndinp = from_random(oplin.jac.target, dtype=oplin.val.dtype)
        assert_allclose(oplin.jac.adjoint(rndinp).extract(varloc.domain),
                        oplin0.jac.adjoint(rndinp), 1e-13, 1e-13)
        foo = oplin.jac.adjoint(rndinp).extract(cstloc.domain)
        assert_equal(foo, 0*foo)

        if isinstance(op, EnergyOperator) and metric_sampling:
            oplin.metric.draw_sample()

        # _jac_vs_finite_differences(op0, varloc, np.sqrt(tol), ntries,
        #                            only_r_differentiable)


def _jac_vs_finite_differences(op, loc, tol, ntries, only_r_differentiable):
    for _ in range(ntries):
        lin = op(Linearization.make_var(loc))
        loc2, lin2 = _get_acceptable_location(op, loc, lin)
        direction = loc2 - loc
        locnext = loc2
        dirnorm = direction.norm()
        hist = []
        for i in range(50):
            locmid = loc + 0.5 * direction
            linmid = op(Linearization.make_var(locmid))
            dirder = linmid.jac(direction)
            numgrad = (lin2.val - lin.val)
            xtol = tol * dirder.norm() / np.sqrt(dirder.size)
            hist.append((numgrad - dirder).norm())
            # print(len(hist),hist[-1])
            if (abs(numgrad - dirder) <= xtol).s_all():
                break
            direction = direction * 0.5
            dirnorm *= 0.5
            loc2, lin2 = locmid, linmid
        else:
            print(hist)
            raise ValueError("gradient and value seem inconsistent")
        loc = locnext
        check_linear_operator(linmid.jac, domain_dtype=loc.dtype,
                              target_dtype=dirder.dtype,
                              only_r_linear=only_r_differentiable,
                              atol=tol**2, rtol=tol**2)


def _check_likelihood_energy(op, loc):
    from .operators.energy_operators import LikelihoodEnergyOperator
    if not isinstance(op, LikelihoodEnergyOperator):
        return
    data_domain = op.data_domain
    if data_domain is None:
        return
    smet = op._sqrt_data_metric_at(loc)
    myassert(smet.domain == smet.target == data_domain)
    nres = op.normalized_residual(loc)
    myassert(nres.domain is data_domain)
    res = op.get_transformation()
    if res is None:
        raise RuntimeError("`get_transformation` is not implemented for "
                            "this LikelihoodEnergyOperator")
    if len(res) != 2:
        raise RuntimeError("`get_transformation` has to return a dtype and the transformation")


def minisanity(likelihood_energy, samples, terminal_colors=True, return_values=False):
    """Log information about the current fit quality and prior compatibility.

    Log a table with fitting information for the likelihood and the prior.
    Assume that the variables in `energy.position.domain` are standard-normal
    distributed a priori. The table contains the reduced chi^2 value, the mean
    and the number of degrees of freedom for every key of a `MultiDomain`. If
    the domain is a `DomainTuple`, the displayed key is `<None>`.

    If everything is consistent the reduced chi^2 values should be close to one
    and the mean of the data residuals close to zero. If the reduced chi^2 value
    in latent space is significantly bigger than one and only one degree of
    freedom is present, the mean column gives an indication in which direction
    to change the respective hyper parameters.

    Ignore all NaN entries in the target of `modeldata_operator` and in `data`.
    Print reduced chi-square values above 2 and 5 in orange and red,
    respectively.

    Parameters
    ----------
    likelihood_energy: LikelihoodEnergyOperator
        Likelihood energy of which the normalized residuals shall be computed.

    samples : SampleListBase
        List of samples.

    terminal_colors : bool, optional
        Setting this to false disables terminal colors. This may be useful if
        the output of minisanity is written to a file. Default: True

    return_values : bool, optional
        If true, in addition to the table in string format, `minisanity` will
        return the computed values as a dictionary. Default: `False`.

    Returns
    -------
        printable_table : string
        values : dictionary
            Only returned if `return_values` is `True`

    Note
    ----
    For computing the reduced chi^2 values and the normalized residuals, the
    metric at `mean` is used.

    """
    from .minimization.sample_list import SampleListBase
    from .sugar import makeDomain

    if not isinstance(samples, SampleListBase):
        raise TypeError(
            "Minisanity takes only SampleLists as input. If you happen to have "
            "only one field (i.e. no samples), you may wrap it via "
            "`ift.SampleList([field])` and pass it to minisanity."
        )

    if not isinstance(likelihood_energy, LikelihoodEnergyOperator):
        return ""

    data_domain = likelihood_energy.data_domain
    latent_domain = samples.domain
    xdoms = [data_domain, latent_domain]

    keylen = 18
    for dom in xdoms:
        if isinstance(dom, MultiDomain):
            keylen = max([max(map(len, dom.keys())), keylen])
    keylen = min([keylen, 42])

    # compute xops
    xops = []
    nres = likelihood_energy.normalized_residual
    if isinstance(data_domain, MultiDomain):
        lam = lambda x: nres(x)
    else:
        name = likelihood_energy.name
        if name is None:
            name = "<None>"
        data_domain = makeDomain({name: data_domain})
        lam = lambda x: nres(x).ducktape_left(name)
    xops.append(lam)
    if isinstance(latent_domain, MultiDomain):
        xops.append(lambda x: x)
    else:
        latent_domain = makeDomain({"<None>": latent_domain})
        xops.append(lambda x: x.ducktape_left("<None>"))
    # /compute xops

    xdoms = [data_domain, latent_domain]
    xredchisq, xscmean, xndof = [], [], []
    for dd in xdoms:
        xredchisq.append({kk: StatCalculator() for kk in dd.keys()})
        xscmean.append({kk: StatCalculator() for kk in dd.keys()})
        xndof.append({})

    for ss1, ss2 in zip(samples.iterator(xops[0]), samples.iterator(xops[1])):
        if isinstance(data_domain, MultiDomain):
            myassert(ss1.domain == data_domain)
        if isinstance(samples.domain, MultiDomain):
            myassert(ss2.domain == samples.domain)
        for ii, ss in enumerate((ss1, ss2)):
            for kk in ss.domain.keys():
                lsize = ss[kk].size - np.sum(np.isnan(ss[kk].val))
                xredchisq[ii][kk].add(np.nansum(abs(ss[kk].val) ** 2) / lsize)
                xscmean[ii][kk].add(np.nanmean(ss[kk].val))
                xndof[ii][kk] = lsize

    for ii in range(2):
        for kk in xredchisq[ii].keys():
            rcs_mean = xredchisq[ii][kk].mean
            sc_mean = xscmean[ii][kk].mean
            try:
                rcs_std = np.sqrt(xredchisq[ii][kk].var)
                sc_std = np.sqrt(xscmean[ii][kk].var)
            except RuntimeError:
                rcs_std = None
                sc_std = None
            xredchisq[ii][kk] = {'mean': rcs_mean, 'std': rcs_std}
            xscmean[ii][kk] = {'mean': sc_mean, 'std': sc_std}

    s0 = _tableentries(xredchisq[0], xscmean[0], xndof[0], keylen, terminal_colors)
    s1 = _tableentries(xredchisq[1], xscmean[1], xndof[1], keylen, terminal_colors)

    n = 38 + keylen
    s = [n * "=",
         (keylen + 2) * " " + "{:>11}".format("reduced χ²")
           + "{:>14}".format("mean") + "{:>11}".format("# dof"),
         n * "-",
         "Data residuals", s0,
         "Latent space", s1,
         n * "="]

    res_string = "\n".join(s)

    if not return_values:
        return res_string
    else:
        res_dict = {
            'redchisq': {
                'data_residuals': xredchisq[0],
                'latent_variables': xredchisq[1]
            },
            'scmean': {
                'data_residuals': xscmean[0],
                'latent_variables': xscmean[1],
            },
            'ndof': {
                'data_residuals': xndof[0],
                'latent_variables': xndof[1]
            }
        }
        return res_string, res_dict


def _tableentries(redchisq, scmean, ndof, keylen, colors):
    class _bcolors:
        WARNING = "\033[33m" if colors else ""
        FAIL = "\033[31m" if colors else ""
        ENDC = "\033[0m" if colors else ""
        BOLD = "\033[1m" if colors else ""

    out = ""
    for kk in redchisq.keys():
        if len(kk) > keylen:
            out += "  " + kk[: keylen - 1] + "…"
        else:
            out += "  " + kk.ljust(keylen)
        foo = f"{redchisq[kk]['mean']:.1f}"
        if redchisq[kk]['std'] is not None:
            foo += f" ± {redchisq[kk]['std']:.1f}"
        if redchisq[kk]['mean'] > 5 or redchisq[kk]['mean'] < 1/5:
            out += _bcolors.FAIL + _bcolors.BOLD + f"{foo:>11}" + _bcolors.ENDC
        elif redchisq[kk]['mean'] > 2 or redchisq[kk]['mean'] < 1/2:
            out += _bcolors.WARNING + _bcolors.BOLD + f"{foo:>11}" + _bcolors.ENDC
        else:
            out += f"{foo:>11}"

        foo = f"{scmean[kk]['mean']:.1f}"
        if scmean[kk]['std'] is not None:
            foo += f" ± {scmean[kk]['std']:.1f}"
        out += f"{foo:>14}"
        out += f"{ndof[kk]:>11}"
        out += "\n"
    return out[:-1]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import partial

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from ..multi_field import MultiField
from .linear_operator import LinearOperator


class OuterProduct(LinearOperator):
    """Performs the point-wise outer product of two fields.

    Parameters
    ---------
    domain : DomainTuple, the domain of the input field
    field : :class:`nifty8.field.Field`
    ---------
    """
    def __init__(self, domain, field):
        self._domain = DomainTuple.make(domain)
        self._field = field
        self._target = DomainTuple.make(
            tuple(sub_d for sub_d in field.domain._dom + self._domain._dom))
        self._capability = self.TIMES | self.ADJOINT_TIMES

        try:
            from jax import numpy as jnp
            from jax.tree_util import tree_map

            from ..re import Field as ReField

            a_j = ReField(field.val) if isinstance(field, (Field, MultiField)) else field

            def jax_expr(x):
                # Preserve the input type
                if not isinstance(x, ReField):
                    a_astype_x = a_j.val if isinstance(a_j, ReField) else a_j
                else:
                    a_astype_x = a_j

                return tree_map(
                    partial(jnp.tensordot, axes=((), ())),
                    a_astype_x, x
                )

            self._jax_expr = jax_expr
        except ImportError:
            self._jax_expr = None

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            return Field(
                self._target, np.multiply.outer(
                    self._field.val, x.val))
        axes = len(self._field.shape)
        return Field(
            self._domain, np.tensordot(self._field.val, x.val, axes))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras, Philipp Frank

import os
import pickle
import re
import time
from warnings import warn

from .. import utilities
from ..field import Field
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..operators.operator import Operator
from ..utilities import get_MPI_params_from_comm, shareRange


class SampleListBase:
    """Base class for storing lists of fields representing samples.

    This class suits as a base class for storing lists of in most cases
    posterior samples. It is intended to be used to hold the minimization state
    of an inference run and comes with a variety of convenience functions like
    computing the mean or standard deviation of the output of a given operator
    over the sample list.

    Parameters
    ----------
    comm : MPI communicator or None
        If not `None`, :class:`SampleListBase` can gather samples across multiple
        MPI tasks. If `None`, :class:`SampleListBase` is not a distributed object.
    domain : Domainoid (can be DomainTuple, MultiDomain, dict, Domain or list of Domains)
        The domain on which the samples are defined.
    n_local_samples : int
        Number of samples present on the current MPI task.

    Note
    ----
    A class inheriting from :class:`SampleListBase` needs to call the constructor of
    `SampleListBase` and needs to implement :attr:`n_local_samples` and :attr:`local_item()`.
    """
    def __init__(self, comm, domain, n_local_samples):
        from ..sugar import makeDomain
        self._comm = comm
        self._domain = makeDomain(domain)
        utilities.check_MPI_equality(self._domain, comm)
        self.local_indices = _compute_local_indices(n_local_samples, comm)

        self._active_comm = comm
        if comm is not None:
            # 0 corresponds to empty tasks
            # 1 corresponds to tasks with samples
            color = 0 if len(self.local_indices) == 0 else 1
            # If the same key is assigned to multiple processes, the conflict is
            # resolved according to the rank in the original communicator
            key = 0
            self._active_comm = comm.Split(color, key)

        self._n_samples = utilities.allreduce_sum([self.n_local_samples], self.comm)

    @property
    def n_local_samples(self):
        """int: Number of local samples."""
        return len(self.local_indices)

    @property
    def n_samples(self):
        """Return number of samples across all MPI tasks."""
        return self._n_samples

    def local_item(self, i):
        """Return ith local sample."""
        raise NotImplementedError

    def local_iterator(self):
        """Return an iterator over all local samples."""
        for i in range(self.n_local_samples):
            yield self.local_item(i)

    @property
    def comm(self):
        """MPI communicator or None: The communicator used for the SampleListBase."""
        return self._comm

    @property
    def MPI_master(self):
        return utilities.get_MPI_params_from_comm(self.comm)[2]

    @property
    def domain(self):
        """DomainTuple or MultiDomain: the domain on which the samples are defined."""
        return self._domain

    def save_to_hdf5(self, file_name, op=None, samples=False, mean=False, std=False,
                     overwrite=False):
        """Write sample list to HDF5 file.

        This function writes sample lists to HDF5 files that contain two
        groups: `samples` and `stats`. `samples` contain the sublabels `0`,
        `1`, ... that number the labels and `stats` contains the sublabels
        `mean` and `standard deviation`. If `self.domain` is an instance of
        :class:`~nifty8.multi_domain.MultiDomain`, these sublabels refer
        themselves to subgroups. For :class:`~nifty8.field.Field`, the sublabel
        refers to an HDF5 data set.

        If quanitities are not requested (e.g. by setting `mean=False`), the
        respective sublabels are not present in the HDF5 file.

        If `op` is an :class:`~nifty8.operators.operator.Operator`, the operator
        string representation, its domain and its target are written to the
        attributes of the HDF5 file.

        Parameters
        ----------
        file_name : str
            File name of output hdf5 file.
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty8.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty8.field.Field` as an input. Default:
            None.
        samples : bool
            If True, samples are written into hdf5 file.
        mean : bool
            If True, mean of samples is written into hdf5 file.
        std : bool
            If True, standard deviation of samples is written into hdf5 file.
        overwrite : bool
            If True, a potentially existing file with the same file name as
            `file_name`, is overwritten.
        """
        import h5py

        if os.path.isfile(file_name):
            if self.MPI_master and overwrite:
                os.remove(file_name)
            if not overwrite:
                raise RuntimeError(f"File {file_name} already exists. Delete it or use "
                                   "`overwrite=True`")
        if not (samples or mean or std):
            raise ValueError("Neither samples nor mean nor standard deviation shall be written.")

        if self.MPI_master:
            f = h5py.File(file_name, "w")
            if isinstance(op, Operator):
                f.attrs["nifty operator string representation"] = str(op)
                f.attrs["nifty operator domain"] = repr(op.domain)
                f.attrs["nifty operator target"] = repr(op.target)
                f.attrs["nifty domain"] = repr(op.target)
            else:
                f.attrs["nifty domain"] = repr(self.domain)
        else:
            f = utilities.Nop()

        # TODO Add some meta information (e.g. k_length for PowerSpace, distances for RGSpace)
        if samples:
            grp = f.create_group("samples")
            for ii, ss in enumerate(self.iterator(op)):
                _field2hdf5(grp, ss, str(ii))
        if std:
            grp = f.create_group("stats")
            m, v = self.sample_stat(op)
        else:
            if mean:
                grp = f.create_group("stats")
                m = self.average(op)
        if mean:
            _field2hdf5(grp, m, "mean")
        if std:
            _field2hdf5(grp, v.sqrt(), "standard deviation")

        f.close()
        _barrier(self.comm)

    def save_to_fits(self, file_name_base, op=None, samples=False, mean=False, std=False,
                     overwrite=False):
        """Write sample list to FITS file.

        This function writes properties of a sample list to a FITS file. This is
        supported if and only if the target of `op` is a two-dimensional
        RGSpace.

        Parameters
        ----------
        file_name_base : str
            File name base of output FITS file, i.e. without `.fits` extension.
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty8.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty8.field.Field` as an input. Default:
            None.
        samples : bool
            If True, samples are written into hdf5 file.
        mean : bool
            If True, mean of samples is written into hdf5 file.
        std : bool
            If True, standard deviation of samples is written into hdf5 file.
        overwrite : bool
            If True, a potentially existing file with the same file name as
            `file_name`, is overwritten.
        """
        # TODO Add support for 3d and 4d fields
        # https://fits.gsfc.nasa.gov/fits_standard.html

        if not (samples or mean or std):
            raise ValueError("Neither samples nor mean nor standard deviation shall be written.")

        if mean or std:
            m, s = self.sample_stat(op)
        if mean:
            self._save_fits_2d(m, file_name_base + "_mean.fits", overwrite)
        if std:
            self._save_fits_2d(s, file_name_base + "_std.fits", overwrite)

        if samples:
            for ii, ss in enumerate(self.iterator(op)):
                self._save_fits_2d(ss, file_name_base + f"_sample_{ii}.fits", overwrite)

    def _save_fits_2d(self, fld, file_name, overwrite):
        import astropy.io.fits as pyfits
        from astropy.time import Time

        from ..domain_tuple import DomainTuple

        dom = fld.domain
        if not isinstance(dom, DomainTuple) or len(dom) != 1 or len(dom[0].shape) != 2:
            raise ValueError("FITS file export is only supported from 2d-fields. "
                             f"Current domain:\n{dom}")
        h = pyfits.Header()
        h["DATE-MAP"] = Time(time.time(), format="unix").iso.split()[0]
        h["CRVAL1"] = h["CRVAL2"] = 0
        h["CRPIX1"] = h["CRPIX2"] = 0
        h["CUNIT1"] = h["CUNIT2"] = "deg"
        h["CDELT1"], h["CDELT2"] = -dom[0].distances[0], dom[0].distances[1]
        h["CTYPE1"] = "RA---SIN"
        h["CTYPE2"] = "DEC---SIN"
        h["EQUINOX"] = 2000

        hdu = pyfits.PrimaryHDU(fld.val[:, :].T, header=h)
        hdulist = pyfits.HDUList([hdu])
        if self.MPI_master:
            hdulist.writeto(file_name, overwrite=overwrite)

    def iterator(self, op=None):
        """Return iterator over all potentially distributed samples.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is returned. Can be an
            :class:`~nifty8.operators.operator.Operator` or any other callable
            that takes a :class:`~nifty8.field.Field` as an input. Default:
            None.

        Note
        ----
        Calling this function involves MPI communication if `comm != None`.
        """
        op = _none_to_id(op)
        if self.comm is not None:
            for itask in range(self.comm.Get_size()):
                for i in range(_bcast(self.n_local_samples, self._comm, itask)):
                    ss = self.local_item(i) if itask == self._comm.Get_rank() else None
                    yield op(_bcast(ss, self._comm, itask))
        else:
            for ss in self.local_iterator():
                yield op(ss)

    def average(self, op=None):
        """Compute average over all potentially distributed samples.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is averaged.


        Note
        ----
        Calling this function involves MPI communication if `comm != None`.

        Note
        ----
        Calling this function involves allocating arrays for all samples and
        averaging them afterwards. If the number of local samples is big and
        `op` is not None, this leads to much temporary memory usage. If the
        output of `op` is just a :class:`~nifty8.field.Field` or
        :class:`~nifty8.multi_field.MultiField`, :attr:`sample_stat()`
        can be used in order to compute the average memory efficiently.
        """
        res = self._prepare_average(op)
        n = self.n_samples
        return utilities.allreduce_sum(res, self.comm) / n

    def _average_tuple(self, op):
        """Compute simultaneous average over all potentially distributed samples.

        Parameters
        ----------
        op : callable
            Callable that is applied to each item in the
            :class:`SampleListBase` before it is averaged. `op` shall return a
            tuple of whose individual averages are computed and returned as
            tuple.

        Note
        ----
        Calling this function involves MPI communication if `comm != None`.
        """
        n = self.n_samples
        res = None
        if len(self.local_indices) > 0:
            res = self._prepare_average(op)
            res = _list_transpose(res)
            res = tuple(utilities.allreduce_sum(rr, self._active_comm) / n for rr in res)
        if self._active_comm is None and self._comm is None:
            return res

        if len(self.local_indices) > 0:
            task_list = self._comm.allgather(self._comm.Get_rank())
        else:
            task_list = self._comm.allgather(-1)
        root = next(filter((-1).__ne__, task_list))
        return self._comm.bcast(res, root=root)

    def _prepare_average(self, op):
        op = _none_to_id(op)
        res = [op(ss) for ss in self.local_iterator()]
        return res

    def sample_stat(self, op=None):
        """Compute mean and variance of samples after applying `op`.

        Parameters
        ----------
        op : callable or None
            Callable that is applied to each item in the :class:`SampleListBase`
            before it is used to compute mean and variance.

        Returns
        -------
        tuple
            A tuple with two items: the mean and the variance.
        """
        from ..probing import StatCalculator
        if self.n_samples == 1:
            res = self.average(op)
            return res, 0*res
        sc = StatCalculator()
        for ss in self.iterator(op):
            sc.add(ss)
        return sc.mean, sc.var

    def save(self, file_name_base, overwrite=False):
        """Serialize SampleList and write it to disk.

        Parameters
        ----------
        file_name_base : str
            File name of the output file without extension. The actual file name
            will have the extension ".pickle" and before that an identifier that
            distunguishes between MPI tasks.
        overwrite : bool
            Existing files are overwritten.

        Note
        ----
        If the instance of :class:`SampleListBase` is distributed, each MPI task
        writes its own file.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, file_name_base, comm=None):
        """Deserialize SampleList from files on disk.

        Parameters
        ----------
        file_name_base : str
            File name of the input file without extension. The actual file name
            will have the extension ".pickle" and before that an identifier that
            distunguishes between MPI tasks.
        comm : MPI communicator or None
            If not `None`, each MPI task reads its own input file.

        Note
        ----
        `file_name_base` needs to be the same string that has been used for
        saving the :class:`SampleListBase`.

        Note
        ----
        The number of MPI tasks used for saving and loading the `SampleList`
        need to be the same.
        """
        raise NotImplementedError

    @classmethod
    def _list_local_sample_files(cls, file_name_base, comm=None):
        """List all sample files that are relevant for the local task.

        All sample files that correspond to `file_name_base` are searched and
        selected based on which rank the local task has.

        Note
        ----
        This function makes sure that the all file numbers between 0 and the
        maximal found number are present. If this is not the case, a
        `RuntimeError` is raised.

        Note
        ----
        This function distributes the samples according to the standard NIFTy
        distribution scheme (see `ift.utilities.shareRange`).
        """
        base_dir, base_file = os.path.split(os.path.abspath(file_name_base))
        files = [ff for ff in os.listdir(base_dir)
                 if re.match(f"{base_file}.[0-9]+.pickle", ff)]
        if len(files) == 0:
            raise RuntimeError(f"No files matching `{file_name_base}.*.pickle`")
        n_samples = max(list(map(lambda x: int(x.split(".")[-2]), files))) + 1

        ntask, rank, _ = get_MPI_params_from_comm(comm)
        local_indices = range(*shareRange(n_samples, ntask, rank))
        files = [f"{file_name_base}.{ii}.pickle" for ii in local_indices]
        for ff in files:
            if not os.path.isfile(ff):
                raise RuntimeError(f"File {ff} not found")
        return files


class ResidualSampleList(SampleListBase):
    def __init__(self, mean, residuals, neg, comm=None):
        """Store samples in terms of a mean and a residual deviation thereof.


        Parameters
        ----------
        mean : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            Mean of the sample list.
        residuals : list of :class:`nifty8.field.Field` or list of :class:`nifty8.multi_field.MultiField`
            List of residuals from the mean. If it is a list of `MultiField`,
            the domain of the residuals can be a subdomain of the domain of
            mean. This results in adding just a zero in respective `MultiField`
            entries.
        neg: list of bool
            This list has to have the same length as `residuals`. If an entry is
            `True`, the respective residual is subtracted and not added.
        comm : MPI communicator or None
            If not `None`, samples can be gathered across multiple MPI tasks. If
            `None`, :class:`ResidualSampleList` is not a distributed object.
        """
        self._m = mean
        self._r = tuple(residuals)
        self._n = tuple(neg)
        super(ResidualSampleList, self).__init__(comm, mean.domain, len(self._r))

        if len(self._r) != len(self._n):
            raise ValueError("Residuals and neg need to have the same length.")
        if any(elem == 0 for elem in utilities.allreduce_sum([[len(self._r)]], comm)):
            warn("ResidualSampleList: (Partially empty) sample list detected.")

        if len(self._r) > 0:
            r_dom = self._r[0].domain
            if not all(rr.domain is r_dom for rr in self._r):
                raise ValueError("All residuals must have the same domain.")
            if isinstance(r_dom, MultiDomain):
                try:
                    self._m.extract(r_dom)
                except:
                    raise ValueError("`residual.domain` must be a subdomain of `mean.domain`.")

        if not all(isinstance(nn, bool) for nn in neg):
            raise TypeError("All entries in neg need to be bool.")

    def local_item(self, i):
        return self._m.flexible_addsub(self._r[i], self._n[i])

    def at(self, mean):
        """Instantiate `ResidualSampleList` with the same residuals as `self`.

        The mean is updated.

        Note
        ----
        If `self.domain` is a :class:`~nifty8.multi_domain.MultiDomain`, the old
        and new mean are combined with
        :attr:`~nifty8.multi_field.MultiField.union` beforehand. This means that
        only the multi field entries present in `mean` are updated.

        Returns
        -------
        ResidualSampleList
            Sample list with updated mean.
        """
        if isinstance(self._m, MultiField) and self.domain is not mean.domain:
            mean = MultiField.union([self._m, mean])
        return ResidualSampleList(mean, self._r, self._n, self.comm)

    def save(self, file_name_base, overwrite=False):
        for ii, isample in enumerate(self.local_indices):
            obj = [self._r[ii], self._n[ii]]
            fname = _sample_file_name(file_name_base, isample)
            _save_to_disk(fname, obj, overwrite)
        if self.MPI_master:
            _save_to_disk(f"{file_name_base}.mean.pickle", self._m, overwrite)
        _barrier(self.comm)

    @classmethod
    def load(cls, file_name_base, comm=None):
        _barrier(comm)
        mean = _load_from_disk(f"{file_name_base}.mean.pickle")
        files = cls._list_local_sample_files(file_name_base, comm)
        tmp = [_load_from_disk(ff) for ff in files]
        res = [aa[0] for aa in tmp]
        neg = [aa[1] for aa in tmp]
        return cls(mean, res, neg, comm=comm)

    @classmethod
    def load_mean(cls, file_name_base):
        return _load_from_disk(f"{file_name_base}.mean.pickle")


class SampleList(SampleListBase):
    def __init__(self, samples, comm=None, domain=None):
        """Store samples as a plain list.

        This is a minimalist implementation of :class:`SampleListBase`. It just
        serves as a (potentially MPI-distributed) wrapper of a list of samples.

        Parameters
        ----------
        samples : list of :class:`nifty8.field.Field` or list of :class:`nifty8.multi_field.MultiField`
            List of samples.
        comm : MPI communicator or None
            If not `None`, samples can be gathered across multiple MPI tasks. If
            `None`, :class:`ResidualSampleList` is not a distributed object.
        domain : DomainTuple, MultiDomain or None
            Sets the domain of the `SampleList`. If `samples` is non-empty and
            `domain` is not None, `domain` has to coincide with the domain of
            the samples. Default: None.
        """
        from ..sugar import makeDomain

        if domain is None:
            if len(samples) == 0:
                raise ValueError("Need to pass `domain` to instantiate empty `SampleList`.")
            else:
                domain = samples[0].domain
        else:
            domain = makeDomain(domain)
        super(SampleList, self).__init__(comm, domain, len(samples))
        self._s = samples
        for ss in self._s:
            utilities.check_object_identity(ss.domain, self.domain)

    def local_item(self, i):
        return self._s[i]

    def save(self, file_name_base, overwrite=False):
        nsample = self.n_samples
        for isample in range(nsample):
            if isample in self.local_indices:
                obj = self._s[isample-self.local_indices[0]]
                fname = _sample_file_name(file_name_base, isample)
                _save_to_disk(fname, obj, overwrite=True)
        _barrier(self.comm)

    @classmethod
    def load(cls, file_name_base, comm=None):
        from ..logger import logger
        _barrier(comm)
        foo = "{file_name_base}.mean.pickle"
        if os.path.isfile(foo):
            logger.warn(f"{foo} is present. Most probably you intended to "
                         "call `ift.ResidualSampleList.load()`.")
        files = cls._list_local_sample_files(file_name_base, comm)
        samples = [_load_from_disk(ff) for ff in files]
        dom = None
        if comm is not None:
            if comm.Get_rank() == 0:
                dom = samples[0].domain
            dom = comm.bcast(dom, 0)
        return cls(samples, comm=comm, domain=dom)


def _none_to_id(obj):
    """If the input is None, replace it with identity map. If input is
    `Operator`, append `force` in order to make sure that the function is
    evaluated if there is a chance that the domains match. Otherwise return
    input.
    """
    if obj is None:
        return lambda x: x
    if isinstance(obj, Operator):
        return lambda x: obj.force(x)
    return obj


def _bcast(obj, comm, root):
    """Broadcast python object from given root.

    Parameters
    ----------
    obj : object
        The object to be broadcasted.
    comm : MPI communicator
        MPI communicator used for the broadcasting.
    root : int
        MPI task number from which the object shall be sent.
    """
    data = obj if comm.Get_rank() == root else None
    return comm.bcast(data, root=root)


def _sample_file_name(file_name_base, isample):
    """Return sample-unique file name.

    This file name that can be used to uniquely write samples potentially from all MPI tasks.

    Parameters
    ----------
    file_name_base : str
        User-defined first part of file name.
    isample : int
        Number of sample
    """
    if not isinstance(isample, int):
        raise TypeError
    return f"{file_name_base}.{isample}.pickle"


def _load_from_disk(file_name):
    with open(file_name, "rb") as f:
        obj = pickle.load(f)
    return obj


def _save_to_disk(file_name, obj, overwrite=False):
    if overwrite and os.path.isfile(file_name):
        os.remove(file_name)
    with open(file_name, "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def _field2hdf5(file_handle, obj, name):
    if not isinstance(name, str):
        raise TypeError
    if isinstance(obj, MultiField):
        grp = file_handle.create_group(name)
        for kk, fld in obj.items():
            _field2hdf5(grp, fld, kk)
        return
    if not isinstance(obj, Field):
        raise TypeError
    file_handle.create_dataset(name, data=obj.val)


def _barrier(comm):
    if comm is not None:
        comm.Barrier()


def _list_transpose(list_of_lists):
    ny = len(list_of_lists[0])
    return [[elem[ii] for elem in list_of_lists] for ii in range(ny)]


def _compute_local_indices(n_local, comm):
    if comm is None:
        return range(n_local)
    n_locals = comm.allgather(n_local)
    start = sum(n_locals[:comm.Get_rank()])
    return range(start, start + n_local)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Philipp Arras
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import importlib
import os
from configparser import ConfigParser
from warnings import warn

from ...utilities import myassert

# FIXME point_estimates, constants?
# FIXME "2**lh0" looks weird. Change syntax? Use "$" for references?
# FIXME Cache operators (especially likelihoods)


class OptimizeKLConfig:
    """Plug config files into ift.optimize_kl.

    If you use this class for the first time, look at an example config file
    (e.g. `demos/getting_started_7_config_file.cfg`) before continuing to read
    the documentation.

    After parsing the config file the following steps are taken:

    - All `base` entries are interpreted. The options that are written out
      explicitly in a given section override what might be imported via a `base`
      entry. `base` inheritance does not work recursively (yet?).
    - Repetitions (e.g. "2*5,3*2") in `optimization.*` are expanded and replaced
      by an explicit list (e.g. "5,5,2,2,2"). The length of the resulting list
      is not allowed to be  longer than `total iterations`. If it is shorter,
      the last value is repeated.
    - All optimization stages (`optimization.*`) are joined into a single stage
      (`optimization.0`). For this, the quantity after the point `.` in
      `optimization.*` is interpreted as `int` and sorted afterwards. This means
      that for example `optimization.02` comes after `optimization.-1`. E.g.
      `optimization.1.1` is not allowed.

    For referring to e.g. likelihood energy operators, the star `*` is used as
    dereferencing operator. The value after the `*` refers to a section in the
    config file that must be instantiable via `OptimizeKLConfig.instantiate_section(key)`.

    Generally, all spaces (` `) in keys are internally replaced by underscores
    (`_`) such that they can be used in function calls.

    Parameters
    ----------
    config_parser : ConfigParser
        ConfigParser that contains all configuration.
    builders : dict
        Dictionary of functions that are used to instantiate e.g. operators.

    Example
    -------
    For an example of a typical config file, look at
    `demos/getting_started_7_config_file.cfg` in the nifty repository.

    Note
    ----
    Not treated by this class: export_operator_outputs, initial_position,
    initial_index, comm, inspect_callback, terminate_callback,
    return_final_position, resume

    Note
    ----
    Make sure that ConfigParser is case-sensitive by setting its attribute
    `optionxform` to `str` before parsing.
    """

    def __init__(self, config_parser, builders):
        if not isinstance(config_parser, ConfigParser):
            raise TypeError
        if config_parser.optionxform != str:
            warn("Consider setting `config_parser.optionxform = str`")
        self._cfg = config_parser
        self._builders = dict(builders)
        self._interpret_base()
        self._interpret_repetitions()
        self._join_optimization_stages()

    @classmethod
    def from_file(cls, file_name, builders):
        """
        Parameters
        ----------
        file_name : str
            File name of the config file that is imported.
        builders : dict
            Dictionary of functions that are used to instantiate e.g. operators.
        """
        cfg = ConfigParser()
        cfg.optionxform = str  # make keys case-sensitive
        if not os.path.isfile(file_name):
            raise RuntimeError(f"`{file_name}` not found")
        cfg.read(file_name)
        return cls(cfg, builders)

    def to_file(self, name):
        """Write configuration in standardized form to file.

        Parameters
        ----------
        name : str
            Path to which the config shall be written.
        """
        with open(name, "w") as f:
            self._cfg.write(f)

    def optimize_kl(self, **kwargs):
        """Do the inference and write the config file to output directory.

        All additional parameters to `ift.optimize_kl` can be passed via
        `kwargs`.
        """
        from ..optimize_kl import optimize_kl

        dct = dict(self)
        os.makedirs(dct["output_directory"], exist_ok=True)
        self.to_file(os.path.join(dct["output_directory"], "optimization.cfg"))
        return optimize_kl(**dct, **kwargs)

    def _interpret_base(self):
        """Replace `base` entry in every section by the content of the section it points to."""
        c = self._cfg
        for section in c:
            if "base" in c[section]:
                base_name = c[section]["base"]

                if base_name not in c:
                    raise RuntimeError(
                        f"the referred section `{base_name}` does not exist"
                    )
                if "base" in c[base_name]:
                    raise RuntimeError("recursive bases not allowed for now")

                # Replace base entry in section by the respective values
                c[section] = {**c[base_name], **c[section]}
                del c[section]["base"]

    def _interpret_repetitions(self):
        """Expand repretitions in sections of the form `optimization.*`.

        For example `2*NewtonCG` expands to `NewtonCG,NewtonCG`.

        If fewer entries than `total iterations` are present, fill up with the
        last value.
        """
        c = self._cfg

        # Only look at sections starting with "optimization."
        for optkey in filter(lambda x: x[:13] == "optimization.", c.keys()):
            sec = c[optkey]
            total_iterations = sec.getint("total iterations")
            for key in filter(lambda x: x != "total iterations", sec):
                if key == "base":
                    raise AssertionError(
                        "`base` must already be interpreted. This is a bug."
                    )

                # Expand multiply "*"
                if "," in sec[key]:
                    # Handle spaces around ","
                    entry_list_pre = map(lambda x: x.strip(), sec[key].split(","))
                    entry_list_post = []
                    for val in entry_list_pre:
                        # Nothing to expand because * not present or dereferencing operator
                        if "*" not in val or val[0] == "*":
                            entry_list_post.append(val)
                            continue

                        # Multiply "*" and dereferencing "*" mixed
                        splt = val.split("**")
                        if len(splt) == 2 and splt[0] != "" and splt[1] != "":
                            fac, val = splt
                            val = "*" + val
                            entry_list_post.extend(int(fac) * [val])
                            continue

                        # actual expansion
                        splt = val.split("*")
                        if len(splt) != 2:
                            raise RuntimeError(
                                f"the expression `{val}` cannot have more than one `*`"
                            )
                        fac, val = splt
                        entry_list_post.extend(int(fac) * [val])
                    sec[key] = ",".join(entry_list_post)

                # Fill up
                entry_list_pre = sec[key].split(",")
                diff = total_iterations - len(entry_list_pre)
                if diff < 0:
                    raise RuntimeError(
                        f"The number of total iterations ({total_iterations}) is at least {-diff} too small."
                    )
                entry_list_post = entry_list_pre + diff * [entry_list_pre[-1]]
                myassert(len(entry_list_post) == total_iterations)
                sec[key] = ",".join(entry_list_post)

    def _join_optimization_stages(self):
        """Join all optimization stages into one stage.

        All sections of the form `optimization.*` are combined into a single
        section called `optimization.0`.
        """
        c = self._cfg
        # Only look at sections starting with "optimization." But this time in ascending order
        lookup = {}
        for optkey in filter(lambda x: x[:13] == "optimization.", c.keys()):
            _, myid = optkey.split(".")
            lookup[int(myid)] = optkey
        optimization_keys = [lookup[kk] for kk in sorted(lookup)]
        # Sorting done.

        # Merge optimization sections together into "optimization.0"
        # Start with first one
        fst_key = optimization_keys[0]
        sec0 = c[fst_key]
        # Add the rest
        for optkey in optimization_keys[1:]:
            sec = c[optkey]

            for key in sec:
                if key == "total iterations":
                    sec0["total iterations"] = str(
                        sec0.getint("total iterations") + sec.getint("total iterations")
                    )
                    continue
                sec0[key] = ",".join([sec0[key], sec[key]])
            # has been merged into sec0 and can be deleted
            del c[optkey]

        # If user has chosen something different than optimization.0 as first stage, normalize it
        if fst_key != "optimization.0":
            tmp = c[fst_key]
            c["optimization.0"] = tmp
            del c[fst_key]

    def _to_callable(self, s, dtype=None):
        """Turn list separated by `,` into function that takes the index and returns the respective entry.

        Additionally all references indicated by `*` are instantiated.
        """

        def f(iteration):
            val = s.split(",")[iteration].strip()
            if val[0] == "*":  # is reference
                val = val[1:]
                val = self.instantiate_section(val)
            if val == "None":
                return None
            if dtype is not None:
                val = dtype(val)
            return val

        return f

    def instantiate_section(self, sec):
        """Instantiate object that is described by a section in the config file.

        There are two mechanisms how this instantiation works:

        - Look up the section key in the `self._builders` dictionary and call
          the respective function.
        - If `custom function` is specified in the section, pass all other
          entries of the section as arguments to the referred function.

        Before the instantiation is performed the inputs are transformed
        according to the type information that is passed in the config file. By
        default all values have type `str`. If `bool`, `float` or `int` shall be
        passed, the syntax `type :: value`, e.g. `float :: 1.2`, needs to be
        used in the config file.
        """
        dct = dict(self._cfg[sec])

        # Instantiate all references
        for kk in dct:
            val = dct[kk]
            if len(val) > 1 and val[0] == "*":  # is reference
                dct[kk] = self.instantiate_section(val[1:])

        # Replace all whitespaces with _
        # FIXME Is here the best place to do this?
        newdct = {}
        for kk, vv in dct.items():
            newdct[kk.replace(" ", "_")] = vv
        dct = newdct

        # Parse dtype
        for kk, vv in dct.items():
            if not isinstance(vv, str):
                continue
            tmp = tuple(map(lambda x: x.strip(), vv.split("::")))
            if len(tmp) == 2:  # type information available
                if tmp[0] == "bool":
                    vv = bool(tmp[1])
                elif tmp[0] == "float":
                    vv = float(tmp[1])
                elif tmp[0] == "int":
                    vv = int(tmp[1])
                elif tmp[0] == "None":
                    vv = None
            dct[kk] = vv

        # Plug into builder or something else
        if sec in self._builders:
            return self._builders[sec](**dct)
        if "custom_function" in dct:
            mod_name, func_name = dct.pop("custom_function").rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            func = getattr(mod, func_name)
            return func(**dct)
        raise RuntimeError(
            f"Provide build routine for `{sec}` in builders dictionary or "
            "reference a `custom_function` in the config file."
        )

    def __iter__(self):
        """Enable conversion to `dict` such that the result of this class can
        easily be passed into `ift.optimize_kl`."""
        cdyn = self._cfg["optimization.0"]

        # static
        copt = self._cfg["optimization"]
        yield "output_directory", copt["output directory"]
        yield "save_strategy", copt["save strategy"]
        yield "plot_energy_history", True
        yield "plot_minisanity_history", True

        # dynamic
        yield "total_iterations", int(cdyn["total iterations"])
        for key in [
            "likelihood_energy",
            "n_samples",
            "transitions",
            "kl_minimizer",
            "sampling_iteration_controller",
            "nonlinear_sampling_minimizer",
        ]:
            key1 = key.replace("_", " ")
            if key == "n_samples":
                yield key, self._to_callable(cdyn[key1], int)
            else:
                yield key, self._to_callable(cdyn[key1])

    def __str__(self):
        s = []
        for key, val in self._cfg.items():
            s += [key]
            s += [f"  {kk}: {vv}" for kk, vv in val.items()]
            s += [""]
        return "\n".join(s)

    def __eq__(self, other):
        for a in "_cfg", "_builders":
            if getattr(self, a) != getattr(other, a):
                return False
        return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2022 Max-Planck-Society
# Authors: Philipp Arras, Philipp Frank
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce

import numpy as np

from .. import random, utilities
from ..linearization import Linearization
from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from ..operators.endomorphic_operator import EndomorphicOperator
from ..operators.energy_operators import GaussianEnergy, StandardHamiltonian
from ..operators.sampling_enabler import SamplingEnabler
from ..operators.sandwich_operator import SandwichOperator
from ..operators.scaling_operator import ScalingOperator
from ..probing import approximation2endo
from ..sugar import makeOp
from ..utilities import get_MPI_params_from_comm, myassert, shareRange
from .descent_minimizers import DescentMinimizer
from .energy import Energy
from .energy_adapter import EnergyAdapter
from .sample_list import ResidualSampleList, SampleListBase


def _reduce_field(field, keys):
    if isinstance(field, MultiField) and len(keys)>0:
        return field.extract_by_keys(set(field.keys()) - set(keys))
    return field


def _reduce_by_keys(field, operator, keys):
    """Partially insert a field into an operator

    If the domain of the operator is an instance of `DomainTuple`

    Parameters
    ----------
    field : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
        Potentially partially constant input field.
    operator : Operator
        Operator into which `field` is partially inserted.
    keys : list
        List of constant `MultiDomain` entries.

    Returns
    -------
    list
        The variable part of the field and the contracted operator.
    """
    from ..sugar import is_fieldlike, is_operator
    myassert(is_fieldlike(field))
    myassert(is_operator(operator))
    if isinstance(field, MultiField):
        cst_field = field.extract_by_keys(keys)
        var_field = field.extract_by_keys(set(field.keys()) - set(keys))
        _, operator = operator.simplify_for_constant_input(cst_field)
        return var_field, operator
    myassert(len(keys) == 0)
    return field, operator


class _SelfAdjointOperatorWrapper(EndomorphicOperator):
    def __init__(self, domain, func):
        from ..sugar import makeDomain
        self._func = func
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._domain = makeDomain(domain)

    def apply(self, x, mode):
        self._check_input(x, mode)
        return self._func(x)


def draw_samples(position, H, minimizer, n_samples, mirror_samples, napprox=0,
                 want_error=False, comm=None):
    if not isinstance(n_samples, int):
        raise TypeError
    if not isinstance(mirror_samples, bool):
        raise TypeError
    if not isinstance(H, StandardHamiltonian):
        raise TypeError
    if isinstance(position, MultiField):
        sam_position = position.extract(H.domain)
    else:
        sam_position = position

    # Construct transformation
    geometric = minimizer is not None
    if geometric:
        tr = H.likelihood_energy.get_transformation()
        if tr is None:
            raise ValueError("Geometric sampling only works for likelihoods")
        dtype, f_lh = tr
        if isinstance(dtype, dict):
            myassert(all([dtype[k] is not None for k in dtype.keys()]))
        else:
            myassert(dtype is not None)
        scale = ScalingOperator(f_lh.target, 1., dtype)
        fl = f_lh(Linearization.make_var(sam_position))

        transformation = ScalingOperator(f_lh.domain, 1.) + fl.jac.adjoint @ f_lh
        transformation_mean = sam_position + fl.jac.adjoint(fl.val)
        # Note: This metric is equivalent to H.metric, except for the case of a
        # `VariableCovarianceGaussianEnergy` with `use_full_fisher = True`.
        met = SamplingEnabler(SandwichOperator.make(fl.jac, scale),
                              ScalingOperator(fl.domain, 1., float),
                              H.iteration_controller)
    else:
        met = H(Linearization.make_var(sam_position, want_metric=True)).metric
    if napprox >= 1:
        met._approximation = makeOp(approximation2endo(met, napprox))
    # /Construct transformation

    # Draw samples
    sseq = random.spawn_sseq(n_samples)
    if mirror_samples:
        sseq = reduce(lambda a, b: a+b, [[ss]*2 for ss in sseq])
    local_samples = []
    local_neg = []
    utilities.check_MPI_synced_random_state(comm)
    utilities.check_MPI_equality(sseq, comm)
    y = None

    ntask, rank, _ = get_MPI_params_from_comm(comm)
    for i in range(*shareRange(len(sseq), ntask, rank)):
        with random.Context(sseq[i]):
            neg = mirror_samples and (i % 2 != 0)
            if not neg or y is None:  # we really need to draw a sample
                y, yi = met.special_draw_sample(True)

            if geometric:
                m = transformation_mean - y if neg else transformation_mean + y
                pos = sam_position - yi if neg else sam_position + yi
                en = GaussianEnergy(m) @ transformation
                en = EnergyAdapter(pos, en, nanisinf=True, want_metric=True)
                en, _ = minimizer(en)
                local_samples.append(en.position - sam_position)
                local_neg.append(False)
            else:
                local_samples.append(yi)
                local_neg.append(neg)
    return ResidualSampleList(position, local_samples, local_neg, comm)


def SampledKLEnergy(position, hamiltonian, n_samples, minimizer_sampling,
                    mirror_samples=True, constants=[], point_estimates=[],
                    napprox=0, comm=None, nanisinf=True):
    """Provides the sampled Kullback-Leibler used for Variational Inference,
    specifically for geometric Variational Inference (geoVI) and Metric
    Gaussian VI (MGVI).

    In geoVI a probability distribution is approximated with a standard
    normal distribution in the canonical coordinate system of the Riemannian
    manifold associated with the metric of the other distribution. The
    coordinate transformation is approximated by expanding around a point.
    The MGVI simplification occurs in case this transformation can be
    approximated using a linear expansion. In order to infer the optimal
    expansion point, a stochastic estimate of the Kullback-Leibler
    divergence is minimized. This estimate is obtained by sampling from the
    approximation using the current expansion point. During minimization
    these samples are kept constant; only the expansion point is updated.
    Due to the typically nonlinear structure of the true distribution these
    samples have to be updated eventually by instantiating a
    `SampledKLEnergy` again. For the true probability distribution the
    standard parametrization is assumed. The samples of this class can be
    distributed among MPI tasks.

    Parameters
    ----------
    position : :class:`nifty8.field.Field`
        Expansion point of the coordinate transformation.
    hamiltonian : :class:`nifty8.operators.energy_operators.StandardHamiltonian`
        Hamiltonian of the approximated probability distribution.
    n_samples : integer
        Number of samples used to stochastically estimate the KL.
    minimizer_samp : DescentMinimizer or None
        Minimizer used to perform the non-linear part of geoVI sampling. If
        it is None, only the linear (MGVI) approximation for sampling is
        used and no further non-linear steps are performed.
    mirror_samples : boolean
        Whether the mirrored version of the drawn samples are also used.
        If true, the number of used samples doubles.
        Mirroring samples stabilizes the KL estimate as extreme
        sample variation is counterbalanced.
        Default is True.
    constants : list
        List of parameter keys that are kept constant during optimization.
        Default is no constants.
    point_estimates : list
        List of parameter keys for which no samples are drawn, but that are
        (possibly) optimized for, corresponding to point estimates of these.
        Default is to draw samples for the complete domain.
    napprox : int
        Number of samples for computing preconditioner for linear sampling.
        No preconditioning is done by default.
    comm : MPI communicator or None
        If not None, samples will be distributed as evenly as possible
        across this communicator. If `mirror_samples` is set, then a sample
        and its mirror image will preferably reside on the same task if
        necessary.
    nanisinf : bool
        If true, nan energies which can happen due to overflows in the
        forward model are interpreted as inf. Thereby, the code does not
        crash on these occasions but rather the minimizer is told that the
        position it has tried is not sensible.

    Note
    ----
    The two lists `constants` and `point_estimates` are independent from
    each other. It is possible to sample along domains which are kept
    constant during minimization and vice versa. If a key is in both lists,
    it will be inserted into the Hamiltonian and removed from the KL.

    Note
    ----
    Mirroring samples can help to stabilize the latent mean as it
    reduces sampling noise. But a mirrored sample involves an additional
    solve of the non-linear part of the transformation. When using MPI, the
    samples get distributed as evenly as possible over all tasks. If the
    number of tasks is smaller then the total number of samples (including
    mirrored ones), the mirrored pairs try to reside on the same task as
    their non mirrored partners. This ensures that at least the linear part
    of the sampling is re-used.

    See also
    --------
    `Geometric Variational Inference`, Philipp Frank, Reimar Leike,
    Torsten A. Enßlin, `<https://arxiv.org/abs/2105.10470>`_
    `<https://doi.org/10.3390/e23070853>`_

    `Metric Gaussian Variational Inference`, Jakob Knollmüller,
    Torsten A. Enßlin, `<https://arxiv.org/abs/1901.11033>`_

    Consider citing these papers, if you use MGVI or geoVI.
    """
    if not isinstance(hamiltonian, StandardHamiltonian):
        raise TypeError
    if hamiltonian.domain is not position.domain:
        raise ValueError
    if not isinstance(n_samples, int):
        raise TypeError
    if not isinstance(mirror_samples, bool):
        raise TypeError
    if not (minimizer_sampling is None or isinstance(minimizer_sampling, DescentMinimizer)):
        raise TypeError
    if isinstance(position, MultiField):
        if not set(constants).issubset(set(position.keys())):
            raise ValueError("Constants are not a subset of the keys of the latent space\n"
                             f"Latent space keys: {position.keys()}\n"
                             f"Constants keys: {constants}")
        if not set(point_estimates).issubset(set(position.keys())):
            raise ValueError("Point estimates are not a subset of the keys of the latent space\n"
                             f"Latent space keys: {position.keys()}\n"
                             f"Point estimate keys: {point_estimates}")
        if set(point_estimates) == set(position.keys()):
            raise RuntimeError('Point estimates for whole domain. Use EnergyAdapter instead.')

    # if comm is not None:
    #     eff_n_samples = (2 if mirror_samples else 1)*n_samples
    #     n_tasks = comm.Get_size()
    #     if n_tasks > eff_n_samples:
    #         raise RuntimeError("More MPI tasks available than number of samples "
    #                            f"({n_tasks} > {eff_n_samples})")

    # If a key is in both lists `constants` and `point_estimates` remove it.
    invariant = list(set(constants).intersection(point_estimates))
    if isinstance(position, MultiField) and len(invariant) > 0:
        inv_pos = position.extract_by_keys(invariant)
    else:
        inv_pos = None
    position, hamiltonian = _reduce_by_keys(position, hamiltonian, invariant)

    _, ham_sampling = _reduce_by_keys(position, hamiltonian, point_estimates)
    sample_list = draw_samples(position, ham_sampling, minimizer_sampling, n_samples,
                               mirror_samples, napprox=napprox, comm=comm)
    return SampledKLEnergyClass(sample_list, hamiltonian, constants, inv_pos, nanisinf)


class SampledKLEnergyClass(Energy):
    """Base class for Energies representing a sampled Kullback-Leibler
    divergence for the variational approximation of a distribution with another
    distribution.

    Supports the samples to be distributed across MPI tasks.
    """
    def __init__(self, sample_list, hamiltonian, constants, invariants, nanisinf):
        myassert(isinstance(sample_list, ResidualSampleList))
        myassert(sample_list.domain is hamiltonian.domain)

        if isinstance(sample_list.domain, MultiDomain):
            if not (invariants is None or isinstance(invariants, MultiField)):
                raise TypeError

        super(SampledKLEnergyClass, self).__init__(_reduce_field(sample_list._m, constants))

        self._sample_list = sample_list
        self._hamiltonian = hamiltonian
        self._nanisinf = bool(nanisinf)
        self._constants = constants
        self._invariants = invariants

        def _func(inp):
            inp, tmp = _reduce_by_keys(inp, hamiltonian, constants)
            tmp = tmp(Linearization.make_var(inp))
            return tmp.val.val[()], tmp.gradient

        self._val, self._grad = sample_list._average_tuple(_func)
        if np.isnan(self._val) and self._nanisinf:
            self._val = np.inf

    @property
    def value(self):
        return self._val

    @property
    def gradient(self):
        return self._grad

    def at(self, position):
        return SampledKLEnergyClass(self._sample_list.at(position),
                                    self._hamiltonian, self._constants,
                                    self._invariants, self._nanisinf)

    def apply_metric(self, x):
        def _func(inp):
            inp, tmp = _reduce_by_keys(inp, self._hamiltonian, self._constants)
            tmp = tmp(Linearization.make_var(inp, want_metric=True))
            return tmp.metric(x)
        return self._sample_list.average(_func)

    @property
    def metric(self):
        return _SelfAdjointOperatorWrapper(self.position.domain,
                                           self.apply_metric)

    @property
    def samples(self):
        if self._invariants is None:
            return self._sample_list
        return self._sample_list.at(self._invariants)



# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from . import refine
from . import refine_util
from . import refine_chart
from . import lanczos
from . import structured_kernel_interpolation
from .conjugate_gradient import cg, static_cg
from .correlated_field import CorrelatedFieldMaker, non_parametric_amplitude
from .energy_operators import (
    Categorical,
    Gaussian,
    Poissonian,
    StudentT,
    VariableCovarianceGaussian,
    VariableCovarianceStudentT,
)
from .field import Field
from .forest_util import (
    ShapeWithDtype,
    assert_arithmetics,
    dot,
    has_arithmetics,
    map_forest,
    map_forest_mean,
    norm,
    shape,
    size,
    stack,
    unite,
    unstack,
    vdot,
    zeros_like,
)
from .hmc import generate_hmc_acc_rej, generate_nuts_tree
from .hmc_oo import HMCChain, NUTSChain
from .kl import (
    GeoMetricKL,
    MetricKL,
    geometrically_sample_standard_hamiltonian,
    mean_hessp,
    mean_metric,
    mean_value_and_grad,
    sample_standard_hamiltonian,
)
from .lanczos import stochastic_lq_logdet
from .likelihood import Likelihood, StandardHamiltonian
from .optimize import minimize, newton_cg, trust_ncg
from .refine_chart import CoordinateChart, RefinementField
from .stats_distributions import (
    invgamma_invprior,
    invgamma_prior,
    laplace_prior,
    lognormal_invprior,
    lognormal_prior,
    normal_prior,
    uniform_prior,
)
from .sugar import (
    ducktape,
    ducktape_left,
    interpolate,
    mean,
    mean_and_std,
    random_like,
    sum_of_squares,
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_equal

from .common import setup_function, teardown_function

pmp = pytest.mark.parametrize


def test_get_signal_variance():
    space = ift.RGSpace(3)
    hspace = space.get_default_codomain()
    sv = ift.get_signal_variance(lambda x: np.ones_like(x), hspace)
    assert_equal(sv, 3.)

    space = ift.RGSpace(3, distances=1.)
    hspace = space.get_default_codomain()

    def spec2(k):
        t = np.zeros_like(k)
        t[k == 0] = 1.
        return t
    assert_equal(ift.get_signal_variance(spec2, hspace), 1/9.)


def test_exec_time():
    dom = ift.RGSpace(12, harmonic=True)
    op = ift.HarmonicTransformOperator(dom)
    op1 = op.ptw("exp")
    lh = ift.GaussianEnergy(domain=op.target, sampling_dtype=np.float64) @ op1
    ic = ift.GradientNormController(iteration_limit=2)
    ham = ift.StandardHamiltonian(lh, ic_samp=ic)
    ham1 = ift.EnergyAdapter(ift.full(ham.domain, 0.), ham)
    kl = ift.SampledKLEnergy(ift.full(ham.domain, 0.), ham, 1, None, mirror_samples=False)
    ops = [op, op1, lh, ham, ham1, kl]
    for oo in ops:
        for wm in [True, False]:
            ift.exec_time(oo, wm)


@pmp('mf', [False, True])
@pmp('cplx', [False, True])
def test_calc_pos(mf, cplx):
    dom = ift.RGSpace(12, harmonic=True)
    op = ift.HarmonicTransformOperator(dom).ptw("exp")
    if mf:
        op = op.ducktape_left('foo')
        dom = ift.makeDomain({'': dom})
    if cplx:
        op = op + 1j*op
    fld = op(0.1 * ift.from_random(op.domain, 'normal'))
    pos = ift.calculate_position(op, fld)
    ift.extra.assert_allclose(op(pos), fld, 1e-1, 1e-1)


def test_isinstance_helpers():
    dom = ift.RGSpace(12, harmonic=True)
    op = ift.ScalingOperator(dom, 12.)
    fld = ift.full(dom, 0.)
    lin = ift.Linearization.make_var(fld)
    assert not ift.is_fieldlike(op)
    assert ift.is_fieldlike(lin)
    assert ift.is_fieldlike(fld)
    assert not ift.is_linearization(op)
    assert ift.is_linearization(lin)
    assert not ift.is_linearization(fld)
    assert ift.is_operator(op)
    assert not ift.is_operator(lin)
    assert not ift.is_operator(fld)

@pmp('dom_shape', [10, (10,20)])
@pmp('n_samples', [2, 5])
@pmp('common_colorbar', (True, False))
def test_plot_priorsamples(dom_shape, n_samples, common_colorbar):
    dom = ift.RGSpace(dom_shape)
    op = ift.ScalingOperator(dom, 1.)
    ift.plot_priorsamples(op, n_samples, common_colorbar, name=f'test_plot_priorsamples_{dom_shape}.png')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .domain_tuple import DomainTuple
from .utilities import check_object_identity, frozendict, indent


class MultiDomain:
    """A tuple of domains corresponding to a direct sum.

    This class is the domain of the direct sum of fields defined on (possibly
    different) domains. To make an instance of this class, call
    `MultiDomain.make(inp)`.

    Notes
    -----
    For consistency and to be independent of the order of insertion, the keys
    within a multi-domain are sorted. Hence, renaming a domain may result in it
    being placed at a different index within a multi-domain. This is especially
    important if a sequence of, e.g., random numbers is distributed sequentially
    over a multi-domain. In this example, ordering keys differently will change
    the resulting :class:`MultiField`.
    """
    _domainCache = {}

    def __init__(self, dct, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError(
                'To create a MultiDomain call `MultiDomain.make()`.')
        self._keys = tuple(sorted(dct.keys()))
        self._domains = tuple(dct[key] for key in self._keys)
        self._idx = frozendict({key: i for i, key in enumerate(self._keys)})

    @staticmethod
    def make(inp):
        """Creates a MultiDomain object from a dictionary of names and domains

        Parameters
        ----------
        inp : MultiDomain or dict{name: DomainTuple}
            The already built MultiDomain or a dictionary of DomainTuples

        Returns
        ------
        A MultiDomain with the input Domains as domains
        """
        if isinstance(inp, MultiDomain):
            return inp
        if not isinstance(inp, dict):
            raise TypeError("dict expected")
        tmp = {}
        for key, value in inp.items():
            if not isinstance(key, str):
                raise TypeError("keys must be strings")
            tmp[key] = DomainTuple.make(value)
        tmp = frozendict(tmp)
        obj = MultiDomain._domainCache.get(tmp)
        if obj is not None:
            return obj
        obj = MultiDomain(tmp, _callingfrommake=True)
        MultiDomain._domainCache[tmp] = obj
        return obj

    def keys(self):
        return self._keys

    def values(self):
        return self._domains

    def domains(self):
        return self._domains

    @property
    def idx(self):
        return self._idx

    def items(self):
        return zip(self._keys, self._domains)

    def __getitem__(self, key):
        return self._domains[self._idx[key]]

    def __len__(self):
        return len(self._keys)

    def __hash__(self):
        return self._keys.__hash__() ^ self._domains.__hash__()

    def __eq__(self, x):
        if self is x:
            return True
        return isinstance(x, MultiDomain) and list(self.items()) == list(x.items())

    def __ne__(self, x):
        return not self.__eq__(x)

    @property
    def size(self):
        return sum(dom.size for dom in self._domains)

    def __str__(self):
        res = "MultiDomain:"
        for key, dom in zip(self._keys, self._domains):
            for ll in f"{key}: {dom}".splitlines():
                res += f"\n  {ll}"
        return res

    @staticmethod
    def union(inp):
        inp = set(inp)
        if len(inp) == 1:  # all domains are identical
            return inp.pop()
        res = {}
        for dom in inp:
            for key, subdom in zip(dom._keys, dom._domains):
                if key in res:
                    check_object_identity(res[key], subdom)
                else:
                    res[key] = subdom
        return MultiDomain.make(res)

    def __reduce__(self):
        return (_unpickleMultiDomain, (dict(self),))

    def __repr__(self):
        return f"MultiDomain.make({dict(self)})"


def _unpickleMultiDomain(*args):
    return MultiDomain.make(*args)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society

import numpy as np
import pytest
from ducc0.misc import vdot

from .common import setup_function, teardown_function

dt = np.float32

def _check(a, b):
    res0 = np.vdot(a, b)
    res1 = vdot(a, b)
    rel_error = np.abs((res0-res1)/(res0+res1)*2)
    assert rel_error < 1e-6


# When this is fixed in numpy, the warning in src/ducc_dispath.py is no longer
# necessary
@pytest.mark.xfail(reason="np.vdot inaccurate for single precision", strict=True)
def test_vdot():
    a = 100*np.ones((1000000,)).astype(dt)
    _check(a, a)


@pytest.mark.xfail(reason="np.vdot inaccurate for single precision", strict=True)
def test_vdot_extreme():
    a = np.array([1e8, 1, -1e8]).astype(dt)
    b = np.array([1e8, 1,  1e8]).astype(dt)
    _check(a, b)



# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from collections.abc import Iterable
from typing import Any, Callable, Dict, Hashable, Mapping, TypeVar, Union

from jax import numpy as jnp
from jax import random
from jax.tree_util import tree_map, tree_reduce, tree_structure, tree_unflatten

from .field import Field

O = TypeVar('O')
I = TypeVar('I')


def isiterable(candidate):
    try:
        iter(candidate)
        return True
    except (TypeError, AttributeError):
        return False


def is1d(ls: Any) -> bool:
    """Indicates whether the input is one dimensional.

    An object is considered one dimensional if it is an iterable of
    non-iterable items.
    """
    if hasattr(ls, "ndim"):
        return ls.ndim == 1
    if not isiterable(ls):
        return False
    return all(not isiterable(e) for e in ls)


def doc_from(original):
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def ducktape(call: Callable[[I], O],
             key: Hashable) -> Callable[[Mapping[Hashable, I]], O]:
    def named_call(p):
        return call(p[key])

    return named_call


def ducktape_left(call: Callable[[I], O],
                  key: Hashable) -> Callable[[I], Dict[Hashable, O]]:
    def named_call(p):
        return {key: call(p)}

    return named_call


def sum_of_squares(tree) -> Union[jnp.ndarray, jnp.inexact]:
    return tree_reduce(jnp.add, tree_map(lambda x: jnp.sum(x**2), tree), 0.)


def mean(forest):
    from functools import reduce

    norm = 1. / len(forest)
    if isinstance(forest[0], Field):
        m = norm * reduce(Field.__add__, forest)
        return m
    else:
        m = norm * reduce(Field.__add__, (Field(t) for t in forest))
        return m.val


def mean_and_std(forest, correct_bias=True):
    if isinstance(forest[0], Field):
        m = mean(forest)
        mean_of_sq = mean(tuple(t**2 for t in forest))
    else:
        m = Field(mean(forest))
        mean_of_sq = Field(mean(tuple(Field(t)**2 for t in forest)))

    n = len(forest)
    scl = jnp.sqrt(n / (n - 1)) if correct_bias else 1.
    std = scl * tree_map(jnp.sqrt, mean_of_sq - m**2)
    if isinstance(forest[0], Field):
        return m, std
    else:
        return m.val, std.val


def random_like(key: Iterable, primals, rng: Callable = random.normal):
    import numpy as np

    struct = tree_structure(primals)
    # Cast the subkeys to the structure of `primals`
    subkeys = tree_unflatten(struct, random.split(key, struct.num_leaves))

    def draw(key, x):
        shp = x.shape if hasattr(x, "shape") else jnp.shape(x)
        dtp = x.dtype if hasattr(x, "dtype") else np.common_type(x)
        return rng(key=key, shape=shp, dtype=dtp)

    return tree_map(draw, subkeys, primals)


def interpolate(xmin=-7., xmax=7., N=14000) -> Callable:
    """Replaces a local nonlinearity such as jnp.exp with a linear interpolation

    Interpolating functions speeds up code and increases numerical stability in
    some cases, but at a cost of precision and range.

    Parameters
    ----------
    xmin : float
        Minimal interpolation value. Default: -7.
    xmax : float
        Maximal interpolation value. Default: 7.
    N : int
        Number of points used for the interpolation. Default: 14000
    """
    def decorator(f):
        from functools import wraps

        x = jnp.linspace(xmin, xmax, N)
        y = f(x)

        @wraps(f)
        def wrapper(t):
            return jnp.interp(t, x, y)

        return wrapper

    return decorator



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from itertools import count

import nifty8 as ift
import numpy as np

from .common import setup_function, teardown_function

name = (f'plot{nr}.png' for nr in count())


def test_plots():
    rg_space1 = ift.makeDomain(ift.RGSpace((10,)))
    rg_space2 = ift.makeDomain(ift.RGSpace((8, 6), distances=1))
    hp_space = ift.makeDomain(ift.HPSpace(5))
    gl_space = ift.makeDomain(ift.GLSpace(10))
    us_space = ift.makeDomain(ift.UnstructuredDomain(10))

    fft = ift.FFTOperator(rg_space2)

    field_rg1_1 = ift.from_random(rg_space1, 'normal')
    field_rg1_2 = ift.from_random(rg_space1, 'normal')
    field_rg2 = ift.from_random(rg_space2, 'normal')
    field_hp = ift.from_random(hp_space, 'normal')
    field_gl = ift.from_random(gl_space, 'normal')
    field_ps = ift.power_analyze(fft.times(field_rg2))
    field_us = ift.from_random(us_space, 'normal')

    plot = ift.Plot()
    plot.add(field_rg1_1, title='Single plot')
    plot.output(name=next(name))

    plot = ift.Plot()
    plot.add(field_rg2, title='2d rg')
    plot.add([field_rg1_1, field_rg1_2], title='list 1d rg', label=['1', '2'])
    plot.add(field_rg1_2, title='1d rg, xmin, ymin', xmin=0.5, ymin=0.,
             xlabel='xmin=0.5', ylabel='ymin=0')
    plot.output(title='Three plots', name=next(name))
    field_us1 = ift.from_random(us_space, 'normal')

    plot = ift.Plot()
    plot.add(field_hp, title='HP planck-color', cmap='Planck-like')
    plot.add(field_rg1_2, title='1d rg')
    plot.add(field_ps)
    plot.add(field_gl, title='GL')
    plot.add(field_rg2, title='2d rg')
    plot.add(field_us, title='Histogram')
    plot.add([field_us, field_us1], label="ii", title='Two Histograms')
    plot.output(title='Seven plots', name=next(name))



def test_mf_plot():
    x_space = ift.RGSpace((16, 16))
    f_space = ift.RGSpace(4)

    d1 = ift.DomainTuple.make([x_space, f_space])
    d2 = ift.DomainTuple.make([f_space, x_space])

    f1 = ift.from_random(d1, 'normal')
    f2 = ift.makeField(d2, np.moveaxis(f1.val, -1, 0))

    plot = ift.Plot()
    plot.add(f1, block=False, title='f_space_idx = 1')
    plot.add(f2, freq_space_idx=0, title='f_space_idx = 0')
    plot.output(nx=2, ny=1, title='MF-Plots, should look identical',
                name=next(name))


def test_iter_plot():
    for ind in range(2):
        x_space = ift.RGSpace((16, 16))
        f_space = ift.UnstructuredDomain(4)

        if ind == 0:
            dom = ift.DomainTuple.make([x_space, f_space])
        else:
            dom = ift.DomainTuple.make([f_space, x_space])

        fld = ift.from_random(dom, 'normal')

        plot = ift.Plot()
        plot.add(fld)
        plot.output(title='Should be 4 plots', name=next(name))


def test_EnergyHistory_plot():
    eh = ift.minimization.iteration_controllers.EnergyHistory()
    for i in range(5):
        eh.append((i, (i + 1)**-2))
    plot = ift.Plot()
    plot.add(eh)
    plot.output(title='EnergyHistory plot', name=next(name))



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Showcasing the Correlated Field model
#
# **Skip to `Parameter Showcases` for the meat/veggies ;)**
#
# The field model roughly works like this:
#
# `f = HT( A * zero_mode * xi ) + offset`
#
# `A` is a spectral power field which is constructed from power spectra that hold on subdomains of the target domain.
# It is scaled by a zero mode operator and then pointwise multiplied by a gaussian excitation field, yielding
# a representation of the field in harmonic space.
# It is then transformed into the target real space and a offset added.
#
# The power spectra `A` is constructed of are in turn constructed as the sum of a power law component
# and an integrated Wiener process whose amplitude and roughness can be set.
#
# ## Setup code

# +
# %matplotlib inline
import nifty8 as ift
import matplotlib.pyplot as plt
plt.rcParams['figure.dpi'] = 100
plt.style.use("seaborn-notebook")
import numpy as np

n_pix = 256
x_space = ift.RGSpace(n_pix)
ift.random.push_sseq_from_seed(1)

# +
# Plotting routine
def plot(fields, spectra, title=None):
    # Plotting preparation is normally handled by nifty8.Plot
    # It is done manually here to be able to tweak details
    # Fields are assumed to have identical domains
    fig = plt.figure(tight_layout=True, figsize=(10, 3))
    if title is not None:
        fig.suptitle(title, fontsize=14)
    
    # Field
    ax1 = fig.add_subplot(1, 2, 1)
    ax1.axhline(y=0., color='k', linestyle='--', alpha=0.25)
    for field in fields:
        dom = field.domain[0]
        xcoord = np.arange(dom.shape[0]) * dom.distances[0]
        ax1.plot(xcoord, field.val)
    ax1.set_xlim(xcoord[0], xcoord[-1])
    ax1.set_ylim(-5., 5.)
    ax1.set_xlabel('x')
    ax1.set_ylabel('f(x)')
    ax1.set_title('Field realizations')
    
    # Spectrum
    ax2 = fig.add_subplot(1, 2, 2)
    for spectrum in spectra:
        xcoord = spectrum.domain[0].k_lengths
        ycoord = spectrum.val_rw()
        ycoord[0] = ycoord[1]
        ax2.plot(xcoord, ycoord)
    ax2.set_ylim(1e-6, 10.)
    ax2.set_xscale('log')
    ax2.set_yscale('log')
    ax2.set_xlabel('k')
    ax2.set_ylabel('p(k)')
    ax2.set_title('Power Spectrum')
    
    fig.align_labels()
    plt.show()

# Helper: draw main sample
main_sample = None
def init_model(m_pars, fl_pars, matern=False):
    global main_sample
    cf = ift.CorrelatedFieldMaker(m_pars["prefix"])
    cf.set_amplitude_total_offset(m_pars["offset_mean"], m_pars["offset_std"])
    cf.add_fluctuations_matern(**fl_pars) if matern else cf.add_fluctuations(**fl_pars)
    field = cf.finalize(prior_info=0)
    main_sample = ift.from_random(field.domain)
    print("model domain keys:", field.domain.keys())
    
# Helper: field and spectrum from parameter dictionaries + plotting
def eval_model(m_pars, fl_pars, title=None, samples=None, matern=False):
    cf = ift.CorrelatedFieldMaker(m_pars["prefix"])
    cf.set_amplitude_total_offset(m_pars["offset_mean"], m_pars["offset_std"])
    cf.add_fluctuations_matern(**fl_pars) if matern else cf.add_fluctuations(**fl_pars)
    field = cf.finalize(prior_info=0)
    spectrum = cf.amplitude
    if samples is None:
        samples = [main_sample]
    field_realizations = [field(s) for s in samples]
    spectrum_realizations = [spectrum.force(s) for s in samples]
    plot(field_realizations, spectrum_realizations, title)

def gen_samples(key_to_vary):
    if key_to_vary is None:
        return [main_sample]
    dct = main_sample.to_dict()
    subdom_to_vary = dct.pop(key_to_vary).domain
    samples = []
    for i in range(8):
        d = dct.copy()
        d[key_to_vary] = ift.from_random(subdom_to_vary)
        samples.append(ift.MultiField.from_dict(d))
    return samples
        
def vary_parameter(parameter_key, values, samples_vary_in=None, matern=False):
    s = gen_samples(samples_vary_in)
    for v in values:
        if parameter_key in cf_make_pars.keys():
            m_pars = {**cf_make_pars, parameter_key: v}
            eval_model(m_pars, cf_x_fluct_pars, f"{parameter_key} = {v}", s, matern)
        else:
            fl_pars = {**cf_x_fluct_pars, parameter_key: v}
            eval_model(cf_make_pars, fl_pars, f"{parameter_key} = {v}", s, matern)


# -

# ## Before the Action: The Moment-Matched Log-Normal Distribution
# Many properties of the correlated field are modelled as being lognormally distributed.
#
# The distribution models are parametrized via their means and standard-deviations (first and second position in tuple).
#
# To get a feeling of how the ratio of the `mean` and `stddev` parameters influences the distribution shape,
# here are a few example histograms: (observe the x-axis!)

# +
fig = plt.figure(figsize=(13, 3.5))
mean = 1.0
sigmas = [1.0, 0.5, 0.1]

for i in range(3):
    op = ift.LognormalTransform(mean=mean, sigma=sigmas[i],
                                key='foo', N_copies=0)
    op_samples = np.array(
        [op(s).val for s in [ift.from_random(op.domain) for i in range(10000)]])

    ax = fig.add_subplot(1, 3, i + 1)
    ax.hist(op_samples, bins=50)
    ax.set_title(f"mean = {mean}, sigma = {sigmas[i]}")
    ax.set_xlabel('x')
    del op_samples

plt.show()
# -

# ## The Neutral Field
#
# To demonstrate the effect of all parameters, first a 'neutral' set of parameters
# is defined which then are varied one by one, showing the effect of the variation
# on the generated field realizations and the underlying power spectrum from which
# they were drawn.
#
# As a neutral field, a model with a white power spectrum and vanishing spectral power was chosen.

# +
# Neutral model parameters yielding a quasi-constant field
cf_make_pars = {
    'offset_mean': 0.,
    'offset_std': (1e-3, 1e-16),
    'prefix': ''
}

cf_x_fluct_pars = {
    'target_subdomain': x_space,
    'fluctuations': (1e-3, 1e-16),
    'flexibility': (1e-3, 1e-16),
    'asperity': (1e-3, 1e-16),
    'loglogavgslope': (0., 1e-16)
}

init_model(cf_make_pars, cf_x_fluct_pars)
# -

# Show neutral field
eval_model(cf_make_pars, cf_x_fluct_pars, "Neutral Field")

# # Parameter Showcases
#
# ## The `fluctuations` parameters of `add_fluctuations()`
#
# determine the **amplitude of variations along the field dimension**
# for which `add_fluctuations` is called.
#
# `fluctuations[0]` set the _average_ amplitude of the fields fluctuations along the given dimension,\
# `fluctuations[1]` sets the width and shape of the amplitude distribution.
#
# The amplitude is modelled as being log-normally distributed,
# see `The Moment-Matched Log-Normal Distribution` above for details.
#
# #### `fluctuations` mean:

vary_parameter('fluctuations', [(0.05, 1e-16), (0.5, 1e-16), (2., 1e-16)], samples_vary_in='xi')

# #### `fluctuations` std:

vary_parameter('fluctuations', [(1., 0.01), (1., 0.1), (1., 1.)], samples_vary_in='fluctuations')
cf_x_fluct_pars['fluctuations'] = (1., 1e-16)

# ## The `loglogavgslope` parameters of `add_fluctuations()`
#
# determine **the slope of the loglog-linear (power law) component of the power spectrum**.
#
# The slope is modelled to be normally distributed.
#
# #### `loglogavgslope` mean:

vary_parameter('loglogavgslope', [(-6., 1e-16), (-2., 1e-16), (2., 1e-16)], samples_vary_in='xi')

# #### `loglogavgslope` std:

vary_parameter('loglogavgslope', [(-2., 0.02), (-2., 0.2), (-2., 2.0)], samples_vary_in='loglogavgslope')
cf_x_fluct_pars['loglogavgslope'] = (-2., 1e-16)

# ## The `flexibility` parameters of `add_fluctuations()`
#
# determine **the amplitude of the integrated Wiener process component of the power spectrum**
# (how strong the power spectrum varies besides the power-law).
#
# `flexibility[0]` sets the _average_ amplitude of the i.g.p. component,\
# `flexibility[1]` sets how much the amplitude can vary.\
# These two parameters feed into a moment-matched log-normal distribution model,
# see above for a demo of its behavior.
#
# #### `flexibility` mean:

vary_parameter('flexibility', [(0.4, 1e-16), (4.0, 1e-16), (12.0, 1e-16)], samples_vary_in='spectrum')

# #### `flexibility` std:

vary_parameter('flexibility', [(4., 0.02), (4., 0.2), (4., 2.)], samples_vary_in='flexibility')
cf_x_fluct_pars['flexibility'] = (4., 1e-16)

# ## The `asperity` parameters of `add_fluctuations()`
#
# `asperity` determines **how rough the integrated Wiener process component of the power spectrum is**.
#
# `asperity[0]` sets the average roughness, `asperity[1]` sets how much the roughness can vary.\
# These two parameters feed into a moment-matched log-normal distribution model,
# see above for a demo of its behavior.
#
# #### `asperity` mean:

vary_parameter('asperity', [(0.001, 1e-16), (1.0, 1e-16), (5., 1e-16)], samples_vary_in='spectrum')

# #### `asperity` std:

vary_parameter('asperity', [(1., 0.01), (1., 0.1), (1., 1.)], samples_vary_in='asperity')
cf_x_fluct_pars['asperity'] = (1., 1e-16)

# ## The `offset_mean` parameter of `CorrelatedFieldMaker()`
#
# The `offset_mean` parameter defines a global additive offset on the field realizations.
#
# If the field is used for a lognormal model `f = field.exp()`, this acts as a global signal magnitude offset.

# Reset model to neutral
cf_x_fluct_pars['fluctuations'] = (1e-3, 1e-16)
cf_x_fluct_pars['flexibility'] = (1e-3, 1e-16)
cf_x_fluct_pars['asperity'] = (1e-3, 1e-16)
cf_x_fluct_pars['loglogavgslope'] = (1e-3, 1e-16)

vary_parameter('offset_mean', [3., 0., -2.])

# ## The `offset_std` parameters of `CorrelatedFieldMaker()`
#
# Variation of the global offset of the field are modelled as being log-normally distributed.
# See `The Moment-Matched Log-Normal Distribution` above for details.
#
# The `offset_std[0]` parameter sets how much NIFTy will vary the offset *on average*.\
# The `offset_std[1]` parameters defines the with and shape of the offset variation distribution.
#
# #### `offset_std` mean:

vary_parameter('offset_std', [(1e-16, 1e-16), (0.5, 1e-16), (2., 1e-16)], samples_vary_in='xi')

# #### `offset_std` std:

vary_parameter('offset_std', [(1., 0.01), (1., 0.1), (1., 1.)], samples_vary_in='zeromode')

# ## Matern fluctuation kernels
#
# The correlated fields model also supports parametrizing the power spectra of field dimensions
# using Matern kernels. In the following, the effects of their parameters are demonstrated.
#
# Contrary to the field fluctuations parametrization showed above, the Matern kernel
# parameters show strong interactions. For example, the field amplitude does not only depend on the
# amplitude scaling parameter `scale`, but on the combination of all three parameters `scale`,
# `cutoff` and `loglogslope`.

# Neutral model parameters yielding a quasi-constant field
cf_make_pars = {
    'offset_mean': 0.,
    'offset_std': (1e-3, 1e-16),
    'prefix': ''
}

cf_x_fluct_pars = {
    'target_subdomain': x_space,
    'scale': (1e-2, 1e-16),
    'cutoff': (1., 1e-16),
    'loglogslope': (-2.0, 1e-16)
}

init_model(cf_make_pars, cf_x_fluct_pars, matern=True)

# Show neutral field
eval_model(cf_make_pars, cf_x_fluct_pars, "Neutral Field", matern=True)

# # Parameter Showcases
#
# ## The `scale` parameters of `add_fluctuations_matern()`
#
# determine the **overall amplitude scaling factor of fluctuations in the target subdomain**
# for which `add_fluctuations_matern` is called.
#
# **It does not set the absolute amplitude**, which depends on all other parameters, too.
#
# `scale[0]` set the _average_ amplitude scaling factor of the fields' fluctuations along the given dimension,\
# `scale[1]` sets the width and shape of the scaling factor distribution.
#
# The scaling factor is modelled as being log-normally distributed,
# see `The Moment-Matched Log-Normal Distribution` above for details.
# 
# #### `scale` mean:

vary_parameter('scale', [(0.01, 1e-16), (0.1, 1e-16), (1.0, 1e-16)], samples_vary_in='xi', matern=True)

# #### `scale` std:

vary_parameter('scale', [(0.5, 0.01), (0.5, 0.1), (0.5, 0.5)], samples_vary_in='scale', matern=True)
cf_x_fluct_pars['scale'] = (0.5, 1e-16)

# ## The `loglogslope` parameters of `add_fluctuations_matern()`
#
# determine **the slope of the loglog-linear (power law) component of the power spectrum**.
#
# `loglogslope[0]` set the _average_ power law exponent of the fields' power spectrum along the given dimension,\
# `loglogslope[1]` sets the width and shape of the exponent distribution.
#
# The `loglogslope` is modelled to be normally distributed.
#
# #### `loglogslope` mean:

vary_parameter('loglogslope', [(-4.0, 1e-16), (-2.0, 1e-16), (-1.0, 1e-16)], samples_vary_in='xi', matern=True)

# As one can see, the field amplitude also depends on the `loglogslope` parameter.
#
# #### `loglogslope` std:

vary_parameter('loglogslope', [(-3., 0.01), (-3., 0.5), (-3., 1.0)], samples_vary_in='loglogslope', matern=True)

# ## The `cutoff` parameters of `add_fluctuations_matern()`
#
# determines **at what wavevector length the power spectrum should transition from constant power
# to following the powerlaw set by `loglogslope`**.
#
# `cutoff[0]` set the _average_ wavevector length at which the power spectrum transition occurs,\
# `cutoff[1]` sets the width and shape of the transition wavevector length distribution.
#
# The cutoff is modelled as being log-normally distributed,
# see `The Moment-Matched Log-Normal Distribution` above for details.
#
# #### `cutoff` mean:

cf_x_fluct_pars['loglogslope'] = (-8.0, 1e-16)
vary_parameter('cutoff', [(1.0, 1e-16), (3.16, 1e-16), (10.0, 1e-16)], samples_vary_in='xi', matern=True)

# #### `cutoff` std:

vary_parameter('cutoff', [(10., 1.0), (10., 3.16), (10., 10.)], samples_vary_in='cutoff', matern=True)
